import pytest
from helpers import for_one_database, for_specific_databases, get_credentials

from turbodbc import DatabaseError, ParameterError, connect
from turbodbc.connect import _make_connection_string
from turbodbc.connection import Connection


def _test_connection_string(expected, actual):
    assert len(expected) == len(actual)
    expected_tokens = expected.split(";")
    actual_tokens = actual.split(";")
    assert len(expected_tokens) == len(actual_tokens)

    for token in expected_tokens:
        assert token in actual_tokens


def test_make_connection_string_with_dsn():
    connection_string = _make_connection_string("my_dsn", user="my_user")
    _test_connection_string(connection_string, "dsn=my_dsn;user=my_user")


def test_make_connection_string_without_dsn():
    connection_string = _make_connection_string(None, user="my_user")
    _test_connection_string(connection_string, "user=my_user")


@for_one_database
def test_connect_returns_connection_when_successful(dsn, configuration):
    connection = connect(dsn, **get_credentials(configuration))
    assert isinstance(connection, Connection)


@for_one_database
def test_connect_returns_connection_with_explicit_dsn(dsn, configuration):
    connection = connect(dsn=dsn, **get_credentials(configuration))
    assert isinstance(connection, Connection)


def test_connect_raises_on_invalid_dsn():
    invalid_dsn = "This data source does not exist"
    with pytest.raises(DatabaseError):
        connect(invalid_dsn)


@for_specific_databases("postgres")
def test_connect_raises_on_invalid_additional_option(dsn, configuration):
    additional_option = {
        configuration["capabilities"]["connection_user_option"]: "invalid user"
    }
    with pytest.raises(DatabaseError):
        connect(dsn=dsn, **additional_option)


def test_connect_raises_on_ambiguous_parameters():
    with pytest.raises(ParameterError):
        connect("foo", connection_string="DRIVER=bar;SERVER=baz;")
    with pytest.raises(ParameterError):
        connect(connection_string="DRIVER=foo;SERVER=bar;", baz="qux")


@for_one_database
def test_connect_with_connection_string(dsn, configuration):
    connection_string = "DSN=%s;" % dsn
    for para, val in get_credentials(configuration).items():
        connection_string = connection_string + f"{para}={val};"  # noqa
    connection = connect(connection_string=connection_string)
    connection.cursor().execute("SELECT 'foo'")
    connection.close()



import pytest
from helpers import for_one_database, get_credentials
from query_fixture import unique_table_name

from turbodbc import DatabaseError, InterfaceError, connect, make_options


@for_one_database
def test_cursor_on_closed_connection_raises(dsn, configuration):
    connection = connect(dsn, **get_credentials(configuration))
    connection.close()

    with pytest.raises(InterfaceError):
        connection.cursor()


@for_one_database
def test_closing_twice_is_ok(dsn, configuration):
    connection = connect(dsn, **get_credentials(configuration))

    connection.close()
    connection.close()


@for_one_database
def test_closing_connection_closes_all_cursors(dsn, configuration):
    connection = connect(dsn, **get_credentials(configuration))
    cursor_1 = connection.cursor()
    cursor_2 = connection.cursor()
    connection.close()

    with pytest.raises(InterfaceError):
        cursor_1.execute("SELECT 42")

    with pytest.raises(InterfaceError):
        cursor_2.execute("SELECT 42")


@for_one_database
def test_no_autocommit(dsn, configuration):
    connection = connect(dsn, **get_credentials(configuration))

    connection.cursor().execute("CREATE TABLE test_no_autocommit (a INTEGER)")
    connection.close()

    connection = connect(dsn, **get_credentials(configuration))
    with pytest.raises(DatabaseError):
        connection.cursor().execute("SELECT * FROM test_no_autocommit")


@for_one_database
def test_commit_on_closed_connection_raises(dsn, configuration):
    connection = connect(dsn, **get_credentials(configuration))
    connection.close()

    with pytest.raises(InterfaceError):
        connection.commit()


@for_one_database
def test_commit(dsn, configuration):
    table_name = unique_table_name()
    connection = connect(dsn, **get_credentials(configuration))

    connection.cursor().execute(f"CREATE TABLE {table_name} (a INTEGER)")
    connection.commit()

    connection.close()

    connection = connect(dsn, **get_credentials(configuration))
    cursor = connection.cursor()
    cursor.execute(f"SELECT * FROM {table_name}")
    results = cursor.fetchall()
    assert results == []

    cursor.execute(f"DROP TABLE {table_name}")
    connection.commit()


@for_one_database
def test_rollback_on_closed_connection_raises(dsn, configuration):
    connection = connect(dsn, **get_credentials(configuration))
    connection.close()

    with pytest.raises(InterfaceError):
        connection.rollback()


@for_one_database
def test_rollback(dsn, configuration):
    table_name = unique_table_name()
    connection = connect(dsn, **get_credentials(configuration))

    connection.cursor().execute(f"CREATE TABLE {table_name} (a INTEGER)")
    connection.rollback()

    with pytest.raises(DatabaseError):
        connection.cursor().execute(f"SELECT * FROM {table_name}")


@for_one_database
def test_autocommit_enabled_at_start(dsn, configuration):
    table_name = unique_table_name()
    options = make_options(autocommit=True)
    connection = connect(
        dsn, turbodbc_options=options, **get_credentials(configuration)
    )

    connection.cursor().execute(f"CREATE TABLE {table_name} (a INTEGER)")
    connection.close()

    connection = connect(dsn, **get_credentials(configuration))
    cursor = connection.cursor()
    cursor.execute(f"SELECT * FROM {table_name}")
    results = cursor.fetchall()
    assert results == []

    cursor.execute(f"DROP TABLE {table_name}")
    connection.commit()


@for_one_database
def test_autocommit_switching(dsn, configuration):
    table_name = unique_table_name()

    connection = connect(dsn, **get_credentials(configuration))
    connection.autocommit = True  # <---
    connection.cursor().execute(f"CREATE TABLE {table_name} (a INTEGER)")
    connection.close()

    options = make_options(autocommit=True)
    connection = connect(
        dsn, turbodbc_options=options, **get_credentials(configuration)
    )
    connection.autocommit = False  # <---
    connection.cursor().execute(f"INSERT INTO {table_name} VALUES (?)", [42])
    connection.close()

    # table is there, but data was not persisted
    connection = connect(dsn, **get_credentials(configuration))
    cursor = connection.cursor()
    cursor.execute(f"SELECT * FROM {table_name}")
    results = cursor.fetchall()
    assert results == []

    cursor.execute(f"DROP TABLE {table_name}")
    connection.commit()


@for_one_database
def test_autocommit_querying(dsn, configuration):
    connection = connect(dsn, **get_credentials(configuration))
    assert connection.autocommit is False
    connection.autocommit = True
    assert connection.autocommit is True
    connection.close()


@for_one_database
def test_pep343_with_statement(dsn, configuration):
    with connect(dsn, **get_credentials(configuration)) as connection:
        cursor = connection.cursor()

    # connection should be closed, test it with the cursor
    with pytest.raises(InterfaceError):
        cursor.execute("SELECT 42")



import datetime

import pytest
from helpers import (
    for_each_database,
    for_each_database_except,
    for_one_database,
    open_cursor,
)
from query_fixture import query_fixture

import turbodbc


def _test_single_row_result_set(configuration, query, expected_row):
    with open_cursor(configuration) as cursor:
        cursor.execute(query)

        if configuration["capabilities"]["supports_row_count"]:
            assert cursor.rowcount == 1
        else:
            assert cursor.rowcount == -1

        row = cursor.fetchone()
        assert row == expected_row

        row = cursor.fetchone()
        assert row is None


@for_each_database
def test_select_with_too_many_parameters_raises(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with pytest.raises(turbodbc.Error):
            cursor.execute("SELECT 42", [42])

        with pytest.raises(turbodbc.Error):
            cursor.executemany("SELECT 42", [[42]])


@for_each_database
def test_select_single_row_NULL_result(dsn, configuration):
    _test_single_row_result_set(configuration, "SELECT NULL", [None])


@for_each_database
def test_select_single_row_integer_result(dsn, configuration):
    _test_single_row_result_set(configuration, "SELECT 42", [42])


@for_each_database
def test_select_single_row_bool_result(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "SELECT TRUE") as query:
            cursor.execute(query)
            row = cursor.fetchone()
            assert row == [True]
        with query_fixture(cursor, configuration, "SELECT FALSE") as query:
            cursor.execute(query)
            row = cursor.fetchone()
            assert row == [False]


@for_each_database
def test_select_single_row_string_result(dsn, configuration):
    _test_single_row_result_set(configuration, "SELECT 'value'", ["value"])


@for_each_database
def test_select_single_row_unicode_result(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "SELECT UNICODE") as query:
            cursor.execute(query)
            row = cursor.fetchone()
            assert row == ["I \u2665 unicode"]


@for_each_database
def test_select_single_row_double_result(configuration, dsn):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "SELECT DOUBLE") as query:
            cursor.execute(query)
            row = cursor.fetchone()
            assert row == [3.14]


@for_each_database
def test_select_single_row_date_result(configuration, dsn):
    _test_single_row_result_set(
        configuration,
        "SELECT CAST('2015-12-31' AS DATE) AS a",
        [datetime.date(2015, 12, 31)],
    )


@for_each_database
def test_select_single_row_timestamp_result(configuration, dsn):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "SELECT TIMESTAMP") as query:
            cursor.execute(query)
            row = cursor.fetchone()
            assert row == [datetime.datetime(2015, 12, 31, 1, 2, 3)]


@for_each_database
def test_select_single_row_large_numeric_result_as_string(configuration, dsn):
    _test_single_row_result_set(
        configuration, "SELECT -1234567890123.123456789", ["-1234567890123.123456789"]
    )


@for_each_database
def test_select_single_row_large_integer_decimal_as_64_bit_type(configuration, dsn):
    with open_cursor(configuration, large_decimals_as_64_bit_types=True) as cursor:
        with query_fixture(
            cursor, configuration, "SELECT LARGE INTEGER DECIMAL"
        ) as query:
            cursor.execute(query)
            row = cursor.fetchone()
            assert row == [42]


@for_each_database
def test_select_single_row_large_fractional_decimal_as_64_bit_type(configuration, dsn):
    with open_cursor(configuration, large_decimals_as_64_bit_types=True) as cursor:
        with query_fixture(
            cursor, configuration, "SELECT LARGE FRACTIONAL DECIMAL"
        ) as query:
            cursor.execute(query)
            row = cursor.fetchone()
            assert row == pytest.approx([3.14])


@for_each_database
def test_select_single_row_multiple_columns(configuration, dsn):
    _test_single_row_result_set(
        configuration, "SELECT 40, 41, 42, 43", [40, 41, 42, 43]
    )


@for_each_database
def test_fetchone(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "SELECT MULTIPLE INTEGERS") as query:
            cursor.execute(query)
            row = cursor.fetchone()
            assert row == [42]
            row = cursor.fetchone()
            assert row == [43]
            row = cursor.fetchone()
            assert row == [44]

            row = cursor.fetchone()
            assert row is None


@for_each_database
def test_fetchall(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "SELECT MULTIPLE INTEGERS") as query:
            cursor.execute(query)
            rows = cursor.fetchall()
            assert len(rows) == 3
            assert rows[0] == [42]
            assert rows[1] == [43]
            assert rows[2] == [44]


@for_each_database
def test_fetchmany_with_default_arraysize(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "SELECT MULTIPLE INTEGERS") as query:
            cursor.execute(query)
            rows = cursor.fetchmany()
            assert len(rows) == 1
            assert rows[0] == [42]


@for_each_database
def test_fetchmany_with_arraysize_parameter(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "SELECT MULTIPLE INTEGERS") as query:
            cursor.execute(query)
            arraysize_parameter = 2

            rows = cursor.fetchmany(arraysize_parameter)
            assert len(rows) == arraysize_parameter
            assert rows[0] == [42]
            assert rows[1] == [43]

            # arraysize exceeds number of remaining rows
            rows = cursor.fetchmany(arraysize_parameter)
            assert len(rows) == 1
            assert rows[0] == [44]


@for_each_database
def test_fetchmany_with_global_arraysize(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "SELECT MULTIPLE INTEGERS") as query:
            cursor.execute(query)

            arraysize_parameter = 2
            cursor.arraysize = arraysize_parameter

            rows = cursor.fetchmany()
            assert len(rows) == arraysize_parameter
            assert rows[0] == [42]
            assert rows[1] == [43]

            # arraysize exceeds number of remaining rows
            rows = cursor.fetchmany()
            assert len(rows) == 1
            assert rows[0] == [44]


@for_each_database
def test_fetchmany_with_bad_arraysize_parameter_raises(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "SELECT MULTIPLE INTEGERS") as query:
            cursor.execute(query)

            with pytest.raises(turbodbc.InterfaceError):
                cursor.fetchmany(-1)
            with pytest.raises(turbodbc.InterfaceError):
                cursor.fetchmany(0)


@for_each_database
def test_fetchmany_with_bad_global_arraysize_raises(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "SELECT MULTIPLE INTEGERS") as query:
            cursor.execute(query)

            cursor.arraysize = -1
            with pytest.raises(turbodbc.InterfaceError):
                cursor.fetchmany()

            cursor.arraysize = 0
            with pytest.raises(turbodbc.InterfaceError):
                cursor.fetchmany()


@for_each_database
def test_number_of_rows_exceeds_buffer_size(dsn, configuration):
    buffer_size = 3
    with open_cursor(configuration, rows_to_buffer=buffer_size) as cursor:
        with query_fixture(cursor, configuration, "INSERT INTEGER") as table_name:
            numbers = buffer_size * 2 + 1
            for i in range(numbers):
                cursor.execute(f"INSERT INTO {table_name} VALUES ({i})")

            cursor.execute(f"SELECT a FROM {table_name}")
            retrieved = cursor.fetchall()
            actual_sum = sum(row[0] for row in retrieved)
            expected_sum = sum(range(numbers))
            assert expected_sum == actual_sum


@for_each_database
def test_description(dsn, configuration):
    capabilities = configuration["capabilities"]

    with open_cursor(configuration) as cursor:
        assert cursor.description is None

        def fix_case(string):
            if capabilities["reports_column_names_as_upper_case"]:
                return string.upper()
            else:
                return string

        with query_fixture(cursor, configuration, "DESCRIPTION") as table_name:
            cursor.execute(f"SELECT * FROM {table_name}")

            nullness_for_null_column = not capabilities["indicates_null_columns"]

            expected = [
                (fix_case("as_int"), turbodbc.NUMBER, None, None, None, None, True),
                (fix_case("as_double"), turbodbc.NUMBER, None, None, None, None, True),
                (fix_case("as_varchar"), turbodbc.STRING, None, None, None, None, True),
                (fix_case("as_date"), turbodbc.DATETIME, None, None, None, None, True),
                (
                    fix_case("as_timestamp"),
                    turbodbc.DATETIME,
                    None,
                    None,
                    None,
                    None,
                    True,
                ),
                (
                    fix_case("as_int_not_null"),
                    turbodbc.NUMBER,
                    None,
                    None,
                    None,
                    None,
                    nullness_for_null_column,
                ),
            ]
            assert expected == cursor.description


@for_each_database_except(["MySQL"])
def test_unicode_column_names(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "UNICODE COLUMN NAME") as table_name:
            cursor.execute(f"SELECT * FROM {table_name}")
            assert cursor.description[0][0] == "I \u2665 Unicode"


@for_one_database
def test_execute_supports_chaining(dsn, configuration):
    with open_cursor(configuration) as cursor:
        rows = cursor.execute("SELECT 42").fetchall()
        assert rows == [[42]]


@for_one_database
def test_executemany_supports_chaining(dsn, configuration):
    with open_cursor(configuration) as cursor:
        rows = cursor.executemany("SELECT 42").fetchall()
        assert rows == [[42]]



from typing import Optional

from turbodbc_intern import connect as intern_connect

from .connection import Connection
from .exceptions import ParameterError, translate_exceptions
from .options import make_options


def _make_connection_string(dsn: Optional[str], **kwargs) -> str:
    if dsn:
        kwargs["dsn"] = dsn
    return ";".join([f"{key}={value}" for key, value in kwargs.items()])


@translate_exceptions
def connect(
    dsn: Optional[str] = None,
    turbodbc_options=None,
    connection_string: Optional[str] = None,
    **kwargs,
) -> Connection:
    r"""
    Create a connection with the database identified by the ``dsn`` or the ``connection_string``.

    :param dsn: Data source name as given in the (unix) odbc.ini file
           or (Windows) ODBC Data Source Administrator tool.
    :param turbodbc_options: Options that control how turbodbc interacts with the database.
           Create such a struct with `turbodbc.make_options()` or leave this blank to take the defaults.
    :param connection_string: Preformatted ODBC connection string.
           Specifying this and dsn or kwargs at the same time raises ParameterError.
    :param \**kwargs: You may specify additional options as you please. These options will go into
           the connection string that identifies the database. Valid options depend on the specific database you
           would like to connect with (e.g. `user` and `password`, or `uid` and `pwd`)
    :return: A connection to your database
    """
    if turbodbc_options is None:
        turbodbc_options = make_options()

    if connection_string is not None and (dsn is not None or len(kwargs) > 0):
        raise ParameterError("Both connection_string and dsn or kwargs specified")

    if connection_string is None:
        connection_string = _make_connection_string(dsn, **kwargs)

    connection = Connection(intern_connect(connection_string, turbodbc_options))

    return connection



#
# turbodbc documentation build configuration file, created by
# sphinx-quickstart on Sun Apr  9 09:48:15 2017.
#
# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

import sphinx_rtd_theme

sys.path.insert(0, os.path.abspath(os.path.join("..", "python")))


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

extensions = ["sphinx.ext.autodoc"]

templates_path = ["_templates"]

source_suffix = ".rst"

master_doc = "index"

project = "turbodbc"
copyright = "2017, Michael König"
author = "Michael König"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = os.environ.get("READTHEDOCS_VERSION", "latest")
# The full version, including alpha/beta/rc tags.
release = version

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# -- Options for autodoc -------------------------------------------------
# see https://docs.readthedocs.io/en/latest/faq.html

try:
    from unittest.mock import MagicMock
except ImportError:
    from unittest.mock import MagicMock


class Mock(MagicMock):
    @classmethod
    def __getattr__(cls, name):
        return MagicMock()


MOCK_MODULES = ["turbodbc_intern"]
sys.modules.update((mod_name, Mock()) for mod_name in MOCK_MODULES)


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "turbodbcdoc"


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {  # type: ignore
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, "turbodbc.tex", "turbodbc Documentation", "Michael König", "manual"),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "turbodbc", "turbodbc Documentation", [author], 1)]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "turbodbc",
        "turbodbc Documentation",
        author,
        "turbodbc",
        "One line description of project.",
        "Miscellaneous",
    ),
]



import datetime

import pyarrow as pa
import pytest
from helpers import (
    for_each_database,
    for_one_database,
    generate_microseconds_with_precision,
    open_cursor,
)
from numpy import array
from numpy.ma import MaskedArray
from query_fixture import query_fixture

column_backends = ["numpy", "arrow"]
for_each_column_backend = pytest.mark.parametrize("column_backend", column_backends)


def _to_columns(values, dtype, column_backend):
    if column_backend == "numpy":
        return [array(values, dtype=dtype)]
    elif column_backend == "arrow":
        columns = pa.Array.from_pandas(array(values, dtype=dtype))
        return pa.Table.from_arrays([columns], ["column"])


def _to_masked_columns(values, dtype, mask, column_backend):
    if column_backend == "numpy":
        return [MaskedArray(values, mask=mask, dtype=dtype)]
    elif column_backend == "arrow":
        columns = pa.array(array(values, dtype=dtype), mask=array(mask))
        return pa.Table.from_arrays([columns], ["column"])


def _test_basic_column(configuration, fixture, values, dtype, column_backend):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, fixture) as table_name:
            columns = _to_columns(values, dtype, column_backend)
            cursor.executemanycolumns(f"INSERT INTO {table_name} VALUES (?)", columns)
            assert cursor.rowcount == len(values)

            results = cursor.execute(
                f"SELECT A FROM {table_name} ORDER BY A"
            ).fetchall()
            assert results == [[value] for value in sorted(values)]


def _test_column_matches_buffer_size(
    configuration, fixture, values, dtype, column_backend
):
    with open_cursor(configuration, parameter_sets_to_buffer=len(values)) as cursor:
        with query_fixture(cursor, configuration, fixture) as table_name:
            columns = _to_columns(values, dtype, column_backend)
            cursor.executemanycolumns(f"INSERT INTO {table_name} VALUES (?)", columns)
            assert cursor.rowcount == len(values)

            results = cursor.execute(
                f"SELECT A FROM {table_name} ORDER BY A"
            ).fetchall()
            assert results == [[value] for value in sorted(values)]


def _test_column_exceeds_buffer_size(
    configuration, fixture, values, dtype, column_backend
):
    with open_cursor(configuration, parameter_sets_to_buffer=2) as cursor:
        with query_fixture(cursor, configuration, fixture) as table_name:
            columns = _to_columns(values, dtype, column_backend)
            cursor.executemanycolumns(f"INSERT INTO {table_name} VALUES (?)", columns)
            assert cursor.rowcount == len(values)

            results = cursor.execute(
                f"SELECT A FROM {table_name} ORDER BY A"
            ).fetchall()
            assert results == [[value] for value in sorted(values)]


def _test_masked_column(configuration, fixture, values, dtype, column_backend):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, fixture) as table_name:
            columns = _to_masked_columns(
                values, dtype, [False, True, False], column_backend
            )
            cursor.executemanycolumns(f"INSERT INTO {table_name} VALUES (?)", columns)
            assert cursor.rowcount == len(values)

            results = cursor.execute(
                f"SELECT A FROM {table_name} ORDER BY A"
            ).fetchall()
            assert results == [[value] for value in [values[0], values[2]]] + [
                [None]
            ] or results == [[None]] + [[value] for value in [values[0], values[2]]]


def _test_masked_column_with_shrunk_mask(configuration, fixture, values, dtype):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, fixture) as table_name:
            columns = [MaskedArray(values, mask=False, dtype=dtype)]
            columns[0].shrink_mask()
            cursor.executemanycolumns(f"INSERT INTO {table_name} VALUES (?)", columns)
            assert cursor.rowcount == len(values)

            results = cursor.execute(
                f"SELECT A FROM {table_name} ORDER BY A"
            ).fetchall()
            assert results == [[value] for value in sorted(values)]


def _test_masked_column_exceeds_buffer_size(
    configuration, fixture, values, dtype, column_backend
):
    with open_cursor(configuration, parameter_sets_to_buffer=2) as cursor:
        with query_fixture(cursor, configuration, fixture) as table_name:
            columns = _to_masked_columns(
                values, dtype, [True, False, True], column_backend
            )
            cursor.executemanycolumns(f"INSERT INTO {table_name} VALUES (?)", columns)
            assert cursor.rowcount == len(values)

            results = cursor.execute(
                f"SELECT A FROM {table_name} ORDER BY A"
            ).fetchall()
            assert results == [[values[1]], [None], [None]] or results == [
                [None],
                [None],
                [values[1]],
            ]


def _test_single_masked_value(configuration, fixture, values, dtype):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, fixture) as table_name:
            columns = [MaskedArray([values[0]], mask=[True], dtype=dtype)]
            cursor.executemanycolumns(f"INSERT INTO {table_name} VALUES (?)", columns)
            assert cursor.rowcount == 1

            results = cursor.execute(
                f"SELECT A FROM {table_name} ORDER BY A"
            ).fetchall()
            assert results == [[None]]


def _full_column_tests(configuration, fixture, values, dtype, column_backend):
    _test_basic_column(configuration, fixture, values, dtype, column_backend)
    _test_column_matches_buffer_size(
        configuration, fixture, values, dtype, column_backend
    )
    _test_column_exceeds_buffer_size(
        configuration, fixture, values, dtype, column_backend
    )
    _test_masked_column(configuration, fixture, values, dtype, column_backend)
    _test_masked_column_exceeds_buffer_size(
        configuration, fixture, values, dtype, column_backend
    )
    if column_backend == "numpy":
        # For these tests there is no equivalent input structure in Python
        _test_masked_column_with_shrunk_mask(configuration, fixture, values, dtype)
        _test_single_masked_value(configuration, fixture, values, dtype)


@for_each_column_backend
@for_each_database
@pytest.mark.parametrize(
    "dtype",
    [
        "int8",
        "int16",
        "int32",
        "int64",
        "uint8",
        "uint16",
        "uint32",
        "uint64",
    ],
)
def test_integer_column(dsn, configuration, column_backend, dtype):
    if column_backend == "numpy" and dtype != "int64":
        pytest.skip("numpy INSERTs only support int64 as integeral dtype")
    elif dtype == "uint64":
        pytest.skip("uint64 values may be too large to fit into int64")
    else:
        _full_column_tests(
            configuration, "INSERT INTEGER", [17, 23, 42], dtype, column_backend
        )


@for_each_column_backend
@for_each_database
def test_float64_column(dsn, configuration, column_backend):
    _full_column_tests(
        configuration, "INSERT DOUBLE", [2.71, 3.14, 6.25], "float64", column_backend
    )


@for_each_column_backend
@for_each_database
def test_datetime64_microseconds_column(dsn, configuration, column_backend):
    supported_digits = configuration["capabilities"]["fractional_second_digits"]
    fractional = generate_microseconds_with_precision(supported_digits)

    _full_column_tests(
        configuration,
        "INSERT TIMESTAMP",
        [
            datetime.datetime(2015, 12, 31, 1, 2, 3, fractional),
            datetime.datetime(2016, 1, 1, 4, 5, 6, fractional),
            datetime.datetime(2017, 5, 6, 7, 8, 9, fractional),
        ],
        "datetime64[us]",
        column_backend,
    )


@for_each_column_backend
@for_each_database
def test_datetime64_nanoseconds_column(dsn, configuration, column_backend):
    supported_digits = configuration["capabilities"]["fractional_second_digits"]
    # C++ unit test checks that conversion method is capable of nanosecond precision
    fractional = generate_microseconds_with_precision(supported_digits)

    _full_column_tests(
        configuration,
        "INSERT TIMESTAMP",
        [
            datetime.datetime(2015, 12, 31, 1, 2, 3, fractional),
            datetime.datetime(2016, 1, 1, 4, 5, 6, fractional),
            datetime.datetime(2017, 5, 6, 7, 8, 9, fractional),
        ],
        "datetime64[ns]",
        column_backend,
    )


@for_each_column_backend
@for_each_database
def test_datetime64_days_column(dsn, configuration, column_backend):
    _full_column_tests(
        configuration,
        "INSERT DATE",
        [
            datetime.date(2015, 12, 31),
            datetime.date(2016, 1, 1),
            datetime.date(2017, 5, 6),
        ],
        "datetime64[D]",
        column_backend,
    )


@for_each_column_backend
@for_each_database
def test_boolean_column(dsn, configuration, column_backend):
    _full_column_tests(
        configuration, "INSERT BOOL", [True, False, True], "bool", column_backend
    )


@for_each_column_backend
@for_each_database
def test_string_column(dsn, configuration, column_backend):
    _full_column_tests(
        configuration,
        "INSERT STRING",
        ["Simple", "Non-unicode", "Strings"],
        "object",
        column_backend,
    )


@for_each_column_backend
@for_each_database
def test_unicode_column(dsn, configuration, column_backend):
    _full_column_tests(
        configuration,
        "INSERT UNICODE",
        ["a\u2665\u2665\u2665\u2665\u2665", "b\u2665", "c\u2665\u2665\u2665"],
        "object",
        column_backend,
    )


def _test_none_in_string_column(configuration, fixture, column_backend):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, fixture) as table_name:
            columns = _to_columns([None], "object", column_backend)
            cursor.executemanycolumns(f"INSERT INTO {table_name} VALUES (?)", columns)

            results = cursor.execute(
                f"SELECT A FROM {table_name} ORDER BY A"
            ).fetchall()
            assert results == [[None]]


@for_each_column_backend
@for_each_database
def test_string_column_with_None(dsn, configuration, column_backend):
    _test_none_in_string_column(configuration, "INSERT STRING", column_backend)


@for_each_column_backend
@for_each_database
def test_unicode_column_with_None(dsn, configuration, column_backend):
    _test_none_in_string_column(configuration, "INSERT UNICODE", column_backend)


@for_each_column_backend
@for_each_database
def test_multiple_columns(dsn, configuration, column_backend):
    with open_cursor(configuration) as cursor:
        with query_fixture(
            cursor, configuration, "INSERT TWO INTEGER COLUMNS"
        ) as table_name:
            columns = [
                array([17, 23, 42], dtype="int64"),
                array([3, 2, 1], dtype="int64"),
            ]
            if column_backend == "arrow":
                columns = [pa.Array.from_pandas(x) for x in columns]
                columns = pa.Table.from_arrays(columns, ["column1", "column2"])
            cursor.executemanycolumns(
                f"INSERT INTO {table_name} VALUES (?, ?)", columns
            )

            results = cursor.execute(
                f"SELECT A, B FROM {table_name} ORDER BY A"
            ).fetchall()
            assert results == [[17, 3], [23, 2], [42, 1]]


@for_one_database
def test_execute_many_columns_creates_result_set(dsn, configuration):
    with open_cursor(configuration) as cursor:
        cursor.executemanycolumns("SELECT 42", [])
        assert cursor.fetchall() == [[42]]


@for_one_database
def test_execute_many_columns_supports_chaining(dsn, configuration):
    with open_cursor(configuration) as cursor:
        rows = cursor.executemanycolumns("SELECT 42", []).fetchall()
        assert rows == [[42]]



from helpers import for_one_database, open_cursor


@for_one_database
def test_multiple_open_connections(dsn, configuration):
    with open_cursor(configuration) as cursor_1:
        assert cursor_1.executemany("SELECT 42").fetchall() == [[42]]

        with open_cursor(configuration) as cursor_2:
            cursor_2.executemany("SELECT 2")
            cursor_1.executemany("SELECT 1")
            assert cursor_2.fetchall() == [[2]]
            assert cursor_1.fetchall() == [[1]]

        assert cursor_1.executemany("SELECT 123").fetchall() == [[123]]



from turbodbc import Rows, make_options


def test_options_without_parameters():
    options = make_options()
    # one of the default parameters tested in the C++ part
    assert options.parameter_sets_to_buffer == 1000


def test_options_with_overrides():
    options = make_options(
        read_buffer_size=Rows(123),
        parameter_sets_to_buffer=2500,
        varchar_max_character_limit=42,
        prefer_unicode=True,
        use_async_io=True,
        autocommit=True,
        large_decimals_as_64_bit_types=True,
        limit_varchar_results_to_max=True,
        force_extra_capacity_for_unicode=True,
        fetch_wchar_as_char=True,
    )

    assert options.read_buffer_size.rows == 123
    assert options.parameter_sets_to_buffer == 2500
    assert options.varchar_max_character_limit == 42
    assert options.prefer_unicode is True
    assert options.use_async_io is True
    assert options.autocommit is True
    assert options.large_decimals_as_64_bit_types is True
    assert options.limit_varchar_results_to_max is True
    assert options.force_extra_capacity_for_unicode is True
    assert options.fetch_wchar_as_char is True



import datetime
from typing import Any, cast

from helpers import (
    for_each_database,
    for_one_database,
    generate_microseconds_with_precision,
    open_cursor,
)
from query_fixture import query_fixture


def _test_insert_many(configuration, fixture_name, data):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, fixture_name) as table_name:
            cursor.executemany(f"INSERT INTO {table_name} VALUES (?)", data)
            assert len(data) == cursor.rowcount
            cursor.execute(f"SELECT a FROM {table_name} ORDER BY a")
            inserted = cursor.fetchall()
            data = [list(row) for row in data]
            assert data == inserted


def _test_insert_one(configuration, fixture_name, data):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, fixture_name) as table_name:
            cursor.execute(f"INSERT INTO {table_name} VALUES (?)", data)
            assert 1 == cursor.rowcount
            cursor.execute(f"SELECT a FROM {table_name}")
            inserted = cursor.fetchall()
            assert [list(data)] == inserted


@for_one_database
def test_execute_with_tuple(dsn, configuration):
    as_tuple = (1,)
    _test_insert_one(configuration, "INSERT INTEGER", as_tuple)


@for_each_database
def test_insert_with_execute(dsn, configuration):
    as_list = [1]
    _test_insert_one(configuration, "INSERT INTEGER", as_list)


@for_each_database
def test_insert_string_column(dsn, configuration):
    _test_insert_many(
        configuration, "INSERT STRING", [["hello"], ["my"], ["test case"]]
    )


@for_each_database
def test_insert_string_max_column(dsn, configuration):
    _test_insert_many(
        configuration, "INSERT STRING MAX", [["hello"], ["my"], ["test case"]]
    )


@for_each_database
def test_insert_string_column_with_truncation(dsn, configuration):
    with open_cursor(
        configuration, varchar_max_character_limit=9, limit_varchar_results_to_max=True
    ) as cursor:
        with query_fixture(cursor, configuration, "INSERT LONG STRING") as table_name:
            cursor.execute(
                f"INSERT INTO {table_name} VALUES (?)",
                ["Truncated strings suck"],
            )
            cursor.execute(f"SELECT a FROM {table_name}")

            assert cursor.fetchall() == [["Truncated"]]


@for_each_database
def test_insert_unicode_column(dsn, configuration):
    _test_insert_many(
        configuration,
        "INSERT UNICODE",
        [["a I \u2665 unicode"], ["b I really d\u00f8"]],
    )


@for_each_database
def test_insert_unicode_max_column(dsn, configuration):
    _test_insert_many(
        configuration,
        "INSERT UNICODE MAX",
        [["a I \u2665 unicode"], ["b I really d\u00f8"]],
    )


@for_each_database
def test_insert_unicode_column_with_truncation(dsn, configuration):
    with open_cursor(
        configuration, varchar_max_character_limit=4, limit_varchar_results_to_max=True
    ) as cursor:
        with query_fixture(cursor, configuration, "INSERT UNICODE MAX") as table_name:
            cursor.execute(
                f"INSERT INTO {table_name} VALUES (?)", ["I \u2665 truncated"]
            )
            cursor.execute(f"SELECT a FROM {table_name}")

            # depending on the database and the settings, this test may cut through the
            # multi-byte UTF-8 representation of \u2665, or it may not if UTF-16 characters
            # are used. Hence, the assertions are more fuzzy than expected.
            truncated = cursor.fetchall()[0]
            assert len(truncated) > 0
            assert len(truncated) <= 4


@for_each_database
def test_insert_bool_column(dsn, configuration):
    _test_insert_many(configuration, "INSERT BOOL", [[False], [True], [True]])


@for_one_database
def test_execute_many_with_tuple(dsn, configuration):
    _test_insert_many(configuration, "INSERT INTEGER", [(1,), (2,), (3,)])


@for_each_database
def test_insert_integer_column(dsn, configuration):
    _test_insert_many(configuration, "INSERT INTEGER", [[1], [2], [3]])


@for_each_database
def test_insert_double_column(dsn, configuration):
    _test_insert_many(configuration, "INSERT DOUBLE", [[1.23], [2.71], [3.14]])


@for_each_database
def test_insert_date_column(dsn, configuration):
    _test_insert_many(
        configuration,
        "INSERT DATE",
        [
            [datetime.date(2015, 12, 31)],
            [datetime.date(2016, 1, 15)],
            [datetime.date(2016, 2, 3)],
        ],
    )


@for_each_database
def test_insert_timestamp_column(dsn, configuration):
    supported_digits = configuration["capabilities"]["fractional_second_digits"]
    fractional = generate_microseconds_with_precision(supported_digits)

    _test_insert_many(
        configuration,
        "INSERT TIMESTAMP",
        [
            [datetime.datetime(2015, 12, 31, 1, 2, 3, fractional)],
            [datetime.datetime(2016, 1, 15, 4, 5, 6, fractional * 2)],
            [datetime.datetime(2016, 2, 3, 7, 8, 9, fractional * 3)],
        ],
    )


@for_each_database
def test_insert_null(dsn, configuration):
    _test_insert_many(configuration, "INSERT INTEGER", [[None]])


@for_each_database
def test_insert_mixed_data_columns(dsn, configuration):
    # second column has mixed data types in the same column
    # first column makes sure values of "good" columns are not affected
    to_insert = [[23, 1.23], [42, 2]]

    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "INSERT MIXED") as table_name:
            cursor.executemany(f"INSERT INTO {table_name} VALUES (?, ?)", to_insert)
            assert len(to_insert) == cursor.rowcount
            cursor.execute(f"SELECT a, b FROM {table_name} ORDER BY a")
            inserted = [list(row) for row in cursor.fetchall()]
            assert to_insert == inserted


@for_each_database
def test_insert_no_parameter_list(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "INSERT INTEGER") as table_name:
            cursor.executemany(f"INSERT INTO {table_name} VALUES (?)")
            assert 0 == cursor.rowcount
            cursor.execute(f"SELECT a FROM {table_name}")
            inserted = [list(row) for row in cursor.fetchall()]
            assert 0 == len(inserted)


@for_each_database
def test_insert_empty_parameter_list(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "INSERT INTEGER") as table_name:
            cursor.executemany(f"INSERT INTO {table_name} VALUES (?)", [])
            assert 0 == cursor.rowcount
            cursor.execute(f"SELECT a FROM {table_name}")
            inserted = [list(row) for row in cursor.fetchall()]
            assert [] == inserted


@for_each_database
def test_insert_number_of_rows_exceeds_buffer_size(dsn, configuration):
    buffer_size = 3
    numbers = buffer_size * 2 + 1
    data = [[i] for i in range(numbers)]

    with open_cursor(configuration, parameter_sets_to_buffer=buffer_size) as cursor:
        with query_fixture(cursor, configuration, "INSERT INTEGER") as table_name:
            cursor.executemany(f"INSERT INTO {table_name} VALUES (?)", data)
            assert len(data) == cursor.rowcount
            cursor.execute(f"SELECT a FROM {table_name} ORDER BY a")
            inserted = [list(row) for row in cursor.fetchall()]
            assert data == inserted


@for_each_database
def test_description_after_insert(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "INSERT INTEGER") as table_name:
            cursor.execute(f"INSERT INTO {table_name} VALUES (42)")
            assert cursor.description is None


@for_each_database
def test_string_with_differing_lengths(dsn, configuration):
    long_strings = [["x" * 5], ["x" * 50], ["x" * 500]]
    # use integer to force rebind to string buffer afterwards
    to_insert = cast(list[list[Any]], [[1]]) + long_strings
    expected = [["1"]] + long_strings

    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "INSERT LONG STRING") as table_name:
            cursor.executemany(f"INSERT INTO {table_name} VALUES (?)", to_insert)
            assert len(to_insert) == cursor.rowcount
            cursor.execute(f"SELECT a FROM {table_name}")
            inserted = [list(row) for row in cursor.fetchall()]
            assert expected == inserted


@for_each_database
def test_rowcount_works_without_parameter_sets(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "INSERT INTEGER") as table_name:
            cursor.execute(f"INSERT INTO {table_name} VALUES (42), (17)")
            assert cursor.rowcount == 2



import pytest
from helpers import for_one_database, get_credentials

from turbodbc import connect

"""
Test optional features mentioned in PEP-249 "behave" as specified
"""


@for_one_database
def test_callproc_unsupported(dsn, configuration):
    cursor = connect(dsn, **get_credentials(configuration)).cursor()

    with pytest.raises(AttributeError):
        cursor.callproc()



import datetime

from turbodbc import Date, Time, Timestamp


def test_constructors_date():
    d = Date(2016, 1, 4)
    assert d == datetime.date(2016, 1, 4)


def test_constructors_time():
    t = Time(1, 2, 3)
    assert t == datetime.time(1, 2, 3)


def test_constructors_timestamp():
    ts = Timestamp(2016, 1, 2, 3, 4, 5)
    assert ts == datetime.datetime(2016, 1, 2, 3, 4, 5)



import json
from datetime import date, datetime

import numpy
import pgdb  # PyGreSQL
import psycopg2
import pyodbc

import turbodbc


def connect(api, dsn):
    if api == "pyodbc":
        return pyodbc.connect(dsn=dsn)
    if api == "PyGreSQL":
        return pgdb.connect(
            database="test_db", host="localhost:5432", user="postgres", password="test"
        )
    if api == "psycopg2":
        return psycopg2.connect(
            "dbname='test_db' user='postgres' host='localhost' password='test'"
        )
    else:
        return turbodbc.connect(
            dsn,
            parameter_sets_to_buffer=100000,
            rows_to_buffer=100000,
            use_async_io=True,
        )


def _column_data(column_type):
    if column_type == "INTEGER":
        return 42
    if column_type == "DOUBLE PRECISION":
        return 3.14
    if column_type == "DATE":
        return date(2016, 1, 2)
    if "VARCHAR" in column_type:
        return "test data"
    if column_type == "TIMESTAMP":
        return datetime(2016, 1, 2, 3, 4, 5)
    raise RuntimeError(f"Unknown column type {column_type}")


def prepare_test_data(cursor, column_types, powers_of_two_lines):
    cursor.execute("DROP TABLE IF EXISTS test_performance")
    columns = [
        f"col{i} {type}" for i, type in zip(range(len(column_types)), column_types)
    ]
    cursor.execute("CREATE TABLE test_performance ({})".format(", ".join(columns)))

    data = [_column_data(type) for type in column_types]
    #     try:
    cursor.execute(
        "INSERT INTO test_performance VALUES ({})".format(
            ", ".join("?" for _ in columns)
        ),
        data,
    )
    #     except:
    #     cursor.execute('INSERT INTO test_performance VALUES ({})'.format(', '.join('%s' for _ in columns)), data)

    for _ in range(powers_of_two_lines):
        cursor.execute("INSERT INTO test_performance SELECT * FROM test_performance")


def _fetchallnumpy(cursor):
    cursor.fetchallnumpy()


def _stream_to_ignore(cursor):
    for _ in cursor:
        pass


def _stream_to_list(cursor):
    [row for row in cursor]


def measure(cursor, extraction_method):
    cursor.execute("SELECT * FROM test_performance")
    start = datetime.now()
    extraction_method(cursor)
    stop = datetime.now()
    return (stop - start).total_seconds()


# if __name__ == "__main__":
powers_of_two = 10
n_rows = 2**powers_of_two
n_runs = 10
column_types = [
    "INTEGER",
    "INTEGER",
    "DOUBLE PRECISION",
    "DOUBLE PRECISION",
    "VARCHAR(20)",
    "DATE",
    "TIMESTAMP",
]
api = "turbodbc"
# extraction_method = _stream_to_ignore
# extraction_method = _stream_to_list
extraction_method = _fetchallnumpy
database = "psql"

connection = connect(api, database)
cursor = connection.cursor()

print(f"Performing benchmark with {n_rows} rows")
prepare_test_data(cursor, column_types, powers_of_two)

runs_list = []
for r in range(n_runs):
    print(f"Run #{r + 1}")
    runs_list.append(measure(cursor, extraction_method))

runs = numpy.array(runs_list)
results = {
    "number_of_runs": n_runs,
    "rows_per_run": n_rows,
    "column_types": column_types,
    "api": api,
    "extraction_method": extraction_method.__name__,
    "database": database,
    "timings": {
        "best": runs.min(),
        "worst": runs.max(),
        "mean": runs.mean(),
        "standard_deviation": runs.std(),
    },
    "rates": {
        "best": n_rows / runs.min(),
        "worst": n_rows / runs.max(),
        "mean": n_rows * numpy.reciprocal(runs).mean(),
        "standard_deviation": n_rows * numpy.reciprocal(runs).std(),
    },
}

print(json.dumps(results, indent=4, separators=(",", ": ")))
file_name = f"results_{database}_{api}{extraction_method.__name__}.json"
with open(file_name, "w") as file:
    json.dump(results, file, indent=4, separators=(",", ": "))

print(f"Wrote results to file {file_name}")



import datetime
from collections import OrderedDict

import numpy
import pytest
from helpers import (
    for_each_database,
    for_one_database,
    generate_microseconds_with_precision,
    open_cursor,
)
from numpy.ma import MaskedArray
from numpy.testing import assert_equal
from query_fixture import query_fixture

import turbodbc


def _fix_case(configuration, string):
    """
    some databases return column names in upper case
    """
    capabilities = configuration["capabilities"]
    if capabilities["reports_column_names_as_upper_case"]:
        return string.upper()
    else:
        return string


@for_one_database
def test_numpy_without_result_set_raises(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with pytest.raises(turbodbc.InterfaceError):
            cursor.fetchallnumpy()


@for_each_database
def test_numpy_empty_column(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "INSERT INTEGER") as table_name:
            cursor.execute(f"SELECT a FROM {table_name}")
            results = cursor.fetchallnumpy()
            assert isinstance(results, OrderedDict)
            assert len(results) == 1  # ncols
            assert isinstance(results[_fix_case(configuration, "a")], MaskedArray)


@for_each_database
def test_numpy_empty_column_batch_fetch(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "INSERT INTEGER") as table_name:
            cursor.execute(f"SELECT a FROM {table_name}")
            batches = cursor.fetchnumpybatches()
            for idx, batch in enumerate(batches):
                assert isinstance(batch, OrderedDict)
                assert len(batch) == 1  # ncols
                assert isinstance(batch[_fix_case(configuration, "a")], MaskedArray)
                assert_equal(len(batch[_fix_case(configuration, "a")]), 0)
            assert_equal(idx, 0)


@for_each_database
def test_numpy_int_column(dsn, configuration):
    with open_cursor(configuration) as cursor:
        cursor.execute("SELECT 42 AS a")
        results = cursor.fetchallnumpy()
        expected = MaskedArray([42], mask=[0])
        assert results[_fix_case(configuration, "a")].dtype == numpy.int64
        assert_equal(results[_fix_case(configuration, "a")], expected)


@for_each_database
def test_numpy_double_column(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "SELECT DOUBLE") as query:
            cursor.execute(query)
            results = cursor.fetchallnumpy()
            expected = MaskedArray([3.14], mask=[0])
            assert results[_fix_case(configuration, "a")].dtype == numpy.float64
            assert_equal(results[_fix_case(configuration, "a")], expected)


@for_each_database
def test_numpy_boolean_column(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "INSERT INDEXED BOOL") as table_name:
            cursor.executemany(
                f"INSERT INTO {table_name} VALUES (?, ?)",
                [[True, 1], [False, 2], [True, 3]],
            )
            cursor.execute(f"SELECT a FROM {table_name} ORDER BY b")
            results = cursor.fetchallnumpy()
            expected = MaskedArray([True, False, True], mask=[0])
            assert results[_fix_case(configuration, "a")].dtype == numpy.bool_
            assert_equal(results[_fix_case(configuration, "a")], expected)


@for_each_database
def test_numpy_binary_column_with_null(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(
            cursor, configuration, "INSERT TWO INTEGER COLUMNS"
        ) as table_name:
            cursor.executemany(
                f"INSERT INTO {table_name} VALUES (?, ?)", [[42, 1], [None, 2]]
            )  # second column to enforce ordering
            cursor.execute(f"SELECT a FROM {table_name} ORDER BY b")
            results = cursor.fetchallnumpy()
            expected = MaskedArray([42, 0], mask=[0, 1])
            assert_equal(results[_fix_case(configuration, "a")], expected)


@for_each_database
def test_numpy_binary_column_larger_than_batch_size(dsn, configuration):
    with open_cursor(configuration, rows_to_buffer=2) as cursor:
        with query_fixture(cursor, configuration, "INSERT INTEGER") as table_name:
            cursor.executemany(
                f"INSERT INTO {table_name} VALUES (?)",
                [[1], [2], [3], [4], [5]],
            )
            cursor.execute(f"SELECT a FROM {table_name} ORDER BY a")
            results = cursor.fetchallnumpy()
            expected = MaskedArray([1, 2, 3, 4, 5], mask=False)
            assert_equal(results[_fix_case(configuration, "a")], expected)


@for_each_database
def test_numpy_batch_fetch(dsn, configuration):
    with open_cursor(configuration, rows_to_buffer=2) as cursor:
        with query_fixture(cursor, configuration, "INSERT INTEGER") as table_name:
            cursor.executemany(
                f"INSERT INTO {table_name} VALUES (?)",
                [[1], [2], [3], [4], [5]],
            )
            cursor.execute(f"SELECT a FROM {table_name} ORDER BY a")
            batches = cursor.fetchnumpybatches()
            expected_batches = [
                MaskedArray([1, 2], mask=False),
                MaskedArray([3, 4], mask=False),
                MaskedArray([5], mask=False),
            ]
            for idx, batch in enumerate(batches):
                expected = expected_batches[idx]
                assert_equal(batch[_fix_case(configuration, "a")], expected)
            assert_equal(idx, 2)


@for_each_database
def test_numpy_timestamp_column(dsn, configuration):
    supported_digits = configuration["capabilities"]["fractional_second_digits"]
    fractional = generate_microseconds_with_precision(supported_digits)
    timestamp = datetime.datetime(2015, 12, 31, 1, 2, 3, fractional)

    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "INSERT TIMESTAMP") as table_name:
            cursor.execute(f"INSERT INTO {table_name} VALUES (?)", [timestamp])
            cursor.execute(f"SELECT a FROM {table_name}")
            results = cursor.fetchallnumpy()
            expected = MaskedArray([timestamp], mask=[0], dtype="datetime64[us]")
            assert results[_fix_case(configuration, "a")].dtype == numpy.dtype(
                "datetime64[us]"
            )
            assert_equal(results[_fix_case(configuration, "a")], expected)


@for_each_database
def test_numpy_date_column(dsn, configuration):
    date = datetime.date(3999, 12, 31)

    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "INSERT DATE") as table_name:
            cursor.execute(f"INSERT INTO {table_name} VALUES (?)", [date])
            cursor.execute(f"SELECT a FROM {table_name}")
            results = cursor.fetchallnumpy()
            expected = MaskedArray([date], mask=[0], dtype="datetime64[D]")
            assert results[_fix_case(configuration, "a")].dtype == numpy.dtype(
                "datetime64[D]"
            )
            assert_equal(results[_fix_case(configuration, "a")], expected)


@for_each_database
def test_numpy_timelike_column_with_null(dsn, configuration):
    fill_value = 0

    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "INSERT TIMESTAMP") as table_name:
            cursor.execute(f"INSERT INTO {table_name} VALUES (?)", [None])
            cursor.execute(f"SELECT a FROM {table_name}")
            results = cursor.fetchallnumpy()
            expected = MaskedArray([42], mask=[1], dtype="datetime64[us]")
            assert_equal(
                results[_fix_case(configuration, "a")].filled(fill_value),
                expected.filled(fill_value),
            )


@for_each_database
def test_numpy_timelike_column_larger_than_batch_size(dsn, configuration):
    timestamps = [
        datetime.datetime(2015, 12, 31, 1, 2, 3),
        datetime.datetime(2016, 1, 5, 4, 5, 6),
        datetime.datetime(2017, 2, 6, 7, 8, 9),
        datetime.datetime(2018, 3, 7, 10, 11, 12),
        datetime.datetime(2019, 4, 8, 13, 14, 15),
    ]

    with open_cursor(configuration, rows_to_buffer=2) as cursor:
        with query_fixture(cursor, configuration, "INSERT TIMESTAMP") as table_name:
            cursor.executemany(
                f"INSERT INTO {table_name} VALUES (?)",
                [[timestamp] for timestamp in timestamps],
            )
            cursor.execute(f"SELECT a FROM {table_name} ORDER BY a")
            results = cursor.fetchallnumpy()
            expected = MaskedArray(timestamps, mask=[0], dtype="datetime64[us]")
            assert_equal(results[_fix_case(configuration, "a")], expected)


@for_each_database
def test_numpy_string_column(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "INSERT STRING") as table_name:
            cursor.execute(f"INSERT INTO {table_name} VALUES (?)", ["this is a test"])
            cursor.execute(f"SELECT a FROM {table_name}")
            results = cursor.fetchallnumpy()
            expected = MaskedArray(["this is a test"], mask=[0], dtype=numpy.object_)
            assert results[_fix_case(configuration, "a")].dtype == numpy.object_
            assert_equal(results[_fix_case(configuration, "a")], expected)


@for_each_database
def test_numpy_string_column_with_truncation(dsn, configuration):
    with open_cursor(
        configuration, varchar_max_character_limit=9, limit_varchar_results_to_max=True
    ) as cursor:
        with query_fixture(cursor, configuration, "INSERT STRING MAX") as table_name:
            cursor.execute(f"INSERT INTO {table_name} VALUES (?)", ["truncated string"])
            cursor.execute(f"SELECT a FROM {table_name}")
            results = cursor.fetchallnumpy()
            expected = MaskedArray(["truncated"], mask=[0], dtype=numpy.object_)
            assert results[_fix_case(configuration, "a")].dtype == numpy.object_
            assert_equal(results[_fix_case(configuration, "a")], expected)


@for_each_database
def test_numpy_unicode_column(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "INSERT UNICODE") as table_name:
            cursor.execute(f"INSERT INTO {table_name} VALUES (?)", ["unicode \u2665"])
            cursor.execute(f"SELECT a FROM {table_name}")
            results = cursor.fetchallnumpy()
            expected = MaskedArray(["unicode \u2665"], mask=[0], dtype=numpy.object_)
            assert results[_fix_case(configuration, "a")].dtype == numpy.object_
            assert_equal(results[_fix_case(configuration, "a")], expected)


@for_each_database
def test_numpy_unicode_column_with_truncation(dsn, configuration):
    with open_cursor(
        configuration,
        rows_to_buffer=1,
        varchar_max_character_limit=9,
        limit_varchar_results_to_max=True,
    ) as cursor:
        with query_fixture(cursor, configuration, "INSERT UNICODE MAX") as table_name:
            cursor.execute(f"INSERT INTO {table_name} VALUES (?)", ["truncated string"])
            cursor.execute(f"SELECT a FROM {table_name}")
            results = cursor.fetchallnumpy()
            expected = MaskedArray(["truncated"], mask=[0], dtype=numpy.object_)
            assert results[_fix_case(configuration, "a")].dtype == numpy.object_
            assert_equal(results[_fix_case(configuration, "a")], expected)


@for_each_database
def test_numpy_string_column_with_null(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "INSERT STRING") as table_name:
            cursor.execute(f"INSERT INTO {table_name} VALUES (?)", [None])
            cursor.execute(f"SELECT a FROM {table_name}")
            results = cursor.fetchallnumpy()
            expected = MaskedArray([None], mask=[0], dtype=numpy.object_)
            assert_equal(results[_fix_case(configuration, "a")], expected)


@for_each_database
def test_numpy_string_column_larger_than_batch_size(dsn, configuration):
    strings = ["abc", "def", "ghi", "jkl", "mno"]
    with open_cursor(configuration, rows_to_buffer=2) as cursor:
        with query_fixture(cursor, configuration, "INSERT STRING") as table_name:
            cursor.executemany(
                f"INSERT INTO {table_name} VALUES (?)",
                [[string] for string in strings],
            )
            cursor.execute(f"SELECT a FROM {table_name} ORDER BY a")
            results = cursor.fetchallnumpy()
            expected = MaskedArray(strings, mask=[0], dtype=numpy.object_)
            assert_equal(results[_fix_case(configuration, "a")], expected)


@for_each_database
def test_numpy_two_columns(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(
            cursor, configuration, "INSERT TWO INTEGER COLUMNS"
        ) as table_name:
            cursor.executemany(
                f"INSERT INTO {table_name} VALUES (?, ?)", [[1, 42], [2, 41]]
            )
            cursor.execute(f"SELECT a, b FROM {table_name} ORDER BY a")
            results = cursor.fetchallnumpy()
            assert_equal(
                results[_fix_case(configuration, "a")], MaskedArray([1, 2], mask=False)
            )
            assert_equal(
                results[_fix_case(configuration, "b")],
                MaskedArray([42, 41], mask=False),
            )



import random
from contextlib import contextmanager


def unique_table_name():
    return f"test_{random.randint(0, 1000000000)}"


@contextmanager
def query_fixture(cursor, configuration, fixture_key):
    """
    Context manager used to set up fixtures for setting up queries.
    :param cursor: This cursor is used to execute queries
    :param configuration: A dictionary containing configuration and fixtures
    :param fixture_key: Identifies the fixture

    The context manager performs the following tasks:
    * If present, create a table or view based on what is written in the
      fixture's "table" or "view" key
    * If present, execute all queries listed in the fixture's "setup" section
    * If present, return the query listed in the fixture's "payload" section;
      else return the name of a temporary table
    * Clean up any tables or views created

    The fixtures dictionary should have the following format:

    {
        "setup": {
            "view": {
                "create": ["CREATE OR REPLACE VIEW {table_name} AS {content}"],
                "drop": ["DROP VIEW {table_name}"]
            },
            "table": {
                "create": ["CREATE OR REPLACE TABLE {table_name} ({content})"],
                "drop": ["DROP TABLE {table_name}"]
            }
        },
        "fixtures": {
            "my_fixture_key_with_a_table": {
                "table": "field1 INTEGER, field2 VARCHAR(10)",
                "setup": ["A POTENTIALLY EMPTY LIST OF ADDITIONAL SQL QUERIES"],
                "payload": "SELECT 42"
            }
            "my_fixture_key_with_a_view": {
                "view": "SELECT 42",
                "setup": ["A POTENTIALLY EMPTY LIST OF ADDITIONAL SQL QUERIES"],
            }

        }
    }

    All sections are optional. Queries may contain
    "{table_name}" to be replaced with a random table (or view) name.
    """
    fixture = configuration["queries"][fixture_key]
    table_name = unique_table_name()

    def _execute_queries(queries, replacements):
        if not isinstance(queries, list):
            queries = [queries]

        for query in queries:
            try:
                cursor.execute(query.format(**replacements))
            except Exception as error:
                raise type(error)(f'Error executing query "{query}": {error}')

    def create_objects():
        if "view" in fixture:
            queries = configuration["setup"]["view"]["create"]
            replacements = {"table_name": table_name, "content": fixture["view"]}
            _execute_queries(queries, replacements)
        if "table" in fixture:
            queries = configuration["setup"]["table"]["create"]
            replacements = {"table_name": table_name, "content": fixture["table"]}
            _execute_queries(queries, replacements)

    def drop_objects():
        if "view" in fixture:
            queries = configuration["setup"]["view"]["drop"]
            replacements = {"table_name": table_name}
            _execute_queries(queries, replacements)
        if "table" in fixture:
            queries = configuration["setup"]["table"]["drop"]
            replacements = {"table_name": table_name}
            _execute_queries(queries, replacements)

    create_objects()
    try:
        if "setup" in fixture:
            replacements = {"table_name": table_name}
            _execute_queries(fixture["setup"], replacements)
        if "payload" in fixture:
            yield fixture["payload"].format(table_name=table_name)
        else:
            yield table_name
    finally:
        drop_objects()



import pyarrow as pa
import pytest
from helpers import for_one_database, open_cursor
from numpy import array
from numpy.ma import MaskedArray
from query_fixture import query_fixture

import turbodbc
from turbodbc import InterfaceError


@for_one_database
def test_column_of_unsupported_type_raises(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "INSERT INTEGER") as table_name:
            columns = ["this is not a NumPy MaskedArray"]
            with pytest.raises(turbodbc.InterfaceError):
                cursor.executemanycolumns(
                    f"INSERT INTO {table_name} VALUES (?)", columns
                )


@for_one_database
def test_columns_of_unequal_sizes_raise(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "INSERT INTEGER") as table_name:
            columns = [
                MaskedArray([1, 2, 3], mask=False, dtype="int64"),
                MaskedArray([1, 2], mask=False, dtype="int64"),
            ]
            with pytest.raises(turbodbc.InterfaceError):
                cursor.executemanycolumns(
                    f"INSERT INTO {table_name} VALUES (?)", columns
                )


@for_one_database
def test_column_with_incompatible_dtype_raises(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "INSERT INTEGER") as table_name:
            columns = [MaskedArray([1, 2, 3], mask=False, dtype="int16")]
            with pytest.raises(turbodbc.InterfaceError):
                cursor.executemanycolumns(
                    f"INSERT INTO {table_name} VALUES (?)", columns
                )


@for_one_database
def test_column_with_multiple_dimensions_raises(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "INSERT INTEGER") as table_name:
            two_dimensional = array([[1, 2, 3], [4, 5, 6]], dtype="int64")
            columns = [two_dimensional]
            with pytest.raises(turbodbc.InterfaceError):
                cursor.executemanycolumns(
                    f"INSERT INTO {table_name} VALUES (?)", columns
                )


@for_one_database
def test_column_with_non_contiguous_data_raises(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "INSERT INTEGER") as table_name:
            two_dimensional = array([[1, 2, 3], [4, 5, 6]], dtype="int64")
            one_dimensional = two_dimensional[:, 1]
            assert not one_dimensional.flags.c_contiguous
            columns = [one_dimensional]
            with pytest.raises(turbodbc.InterfaceError):
                cursor.executemanycolumns(
                    f"INSERT INTO {table_name} VALUES (?)", columns
                )


@for_one_database
def test_number_of_columns_does_not_match_parameter_count(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "INSERT INTEGER") as table_name:
            columns = [array([42], dtype="int64"), array([17], dtype="int64")]
            with pytest.raises(turbodbc.InterfaceError):
                cursor.executemanycolumns(
                    f"INSERT INTO {table_name} VALUES (?)", columns
                )


@for_one_database
def test_passing_empty_list_of_columns_is_ok(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "INSERT INTEGER") as table_name:
            cursor.executemanycolumns(f"INSERT INTO {table_name} VALUES (42)", [])

            results = cursor.execute(
                f"SELECT A FROM {table_name} ORDER BY A"
            ).fetchall()
            assert results == [[42]]


@for_one_database
def test_passing_empty_column_is_ok(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "INSERT INTEGER") as table_name:
            columns = [array([], dtype="int64")]
            cursor.executemanycolumns(f"INSERT INTO {table_name} VALUES (?)", columns)

            results = cursor.execute(
                f"SELECT A FROM {table_name} ORDER BY A"
            ).fetchall()
            assert results == []


@for_one_database
def test_arrow_table_exceeds_expected_columns(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(
            cursor, configuration, "INSERT TWO INTEGER COLUMNS"
        ) as table_name:
            columns = [
                array([17, 23, 42], dtype="int64"),
                array([3, 2, 1], dtype="int64"),
                array([17, 23, 42], dtype="int64"),
            ]
            columns = [pa.Array.from_pandas(x) for x in columns]
            columns = pa.Table.from_arrays(columns, ["column1", "column2", "column3"])
            # InterfaceError: Number of passed columns (3) is not equal to the number of parameters (2)
            with pytest.raises(InterfaceError):
                cursor.executemanycolumns(
                    f"INSERT INTO {table_name} VALUES (?, ?)", columns
                )


@for_one_database
def test_arrow_table_chunked_arrays_not_supported(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "INSERT INTEGER") as table_name:
            arr = pa.array([1, 2])
            rb = pa.RecordBatch.from_arrays([arr], ["a"])
            table = pa.Table.from_batches([rb, rb])
            with pytest.raises(NotImplementedError):
                cursor.executemanycolumns(f"INSERT INTO {table_name} VALUES (?)", table)



import pytest
from helpers import for_one_database  # , open_cursor

# from query_fixture import query_fixture


@for_one_database
def test_many_batches_with_async_io(dsn, configuration):
    pytest.skip("async_io will be removed in future")
    # with open_cursor(configuration, use_async_io=True) as cursor:
    #    with query_fixture(cursor, configuration, "INSERT INTEGER") as table_name:
    #        # insert 2^16 rows
    #        cursor.execute(f"INSERT INTO {table_name} VALUES (1)")
    #        for _ in range(16):
    #            cursor.execute(f"INSERT INTO {table_name} SELECT * FROM {table_name}")

    #        cursor.execute(f"SELECT * FROM {table_name}")
    #        assert sum(1 for _ in cursor) == 2**16



from helpers import (
    for_each_database_except,
    for_one_database,
    for_specific_databases,
    get_credentials,
)

from turbodbc import connect


@for_one_database
def test_nextset_supported(dsn, configuration):
    cursor = connect(dsn, **get_credentials(configuration)).cursor()
    assert "nextset" in dir(cursor)


@for_one_database
def test_nextset_with_one_result_set(dsn, configuration):
    cursor = connect(dsn, **get_credentials(configuration)).cursor()
    cursor.execute("SELECT 42")
    try:
        cursor.nextset()
    except Exception as exc:
        assert False, f"Didn't find a call for nextset\n{exc}\n"
    else:
        assert True, "Found call for nextset"


@for_specific_databases("postgres")
def test_nextset_with_function(dsn, configuration):
    cursor = connect(dsn, **get_credentials(configuration)).cursor()
    multi_result_set_func = """CREATE FUNCTION TEST_FUNC ()
    RETURNS SETOF int AS
    $func$
    BEGIN
    RETURN QUERY SELECT 4;
    RETURN QUERY SELECT 2;
    END
    $func$ LANGUAGE plpgsql;
    """
    cursor.execute(multi_result_set_func)
    cursor.execute("select TEST_FUNC();")
    try:
        assert cursor.fetchall() == [[4], [2]]
    except Exception as exc:
        assert False, f"Didn't find a call for nextset\n{exc}\n"
    else:
        assert True, "Found call for nextset"


@for_specific_databases("postgres")
def test_nextset_with_postgres_procedure(dsn, configuration):
    cursor = connect(dsn, **get_credentials(configuration)).cursor()
    multi_result_set_func = """CREATE PROCEDURE TEST_PROC(INOUT _result_one
    REFCURSOR = 'result_one', INOUT _result_two REFCURSOR = 'result_two')
    LANGUAGE 'plpgsql'
    AS $BODY$ BEGIN
        OPEN _result_one FOR SELECT 4;

        OPEN _result_two FOR SELECT 2;
    END $BODY$;
    """
    cursor.execute(multi_result_set_func)
    multi_result_call_func = """
    CALL TEST_PROC();
    BEGIN;
        FETCH ALL FROM "result_one";
        FETCH ALL FROM "result_two";
    COMMIT;
    """
    cursor.execute(multi_result_call_func)
    try:
        assert cursor.fetchall() == [["result_one", "result_two"]]
    except Exception as exc:
        assert False, f"Didn't find a call for nextset\n{exc}\n"
    else:
        assert True, "Found call for nextset"


@for_specific_databases("postgres")
def test_nextset_with_postgres_function(dsn, configuration):
    cursor = connect(dsn, **get_credentials(configuration)).cursor()
    multi_result_set_func = """CREATE FUNCTION TEST_FUNC() RETURNS SETOF refcursor as $$
    DECLARE
        result_one refcursor := 'result_one';
        result_two refcursor := 'result_two';
    BEGIN
        OPEN result_one FOR SELECT 4;
        RETURN NEXT result_one;
        OPEN result_two FOR SELECT 2;
        RETURN NEXT result_two;
    END;
    $$ LANGUAGE plpgsql;
    """
    cursor.execute(multi_result_set_func)
    multi_result_call_func = """
    select * from TEST_FUNC();
    BEGIN;
        FETCH ALL FROM "result_one";
        FETCH ALL FROM "result_two";
    COMMIT;
    """
    cursor.execute(multi_result_call_func)
    try:
        assert cursor.fetchall() == [["result_one"], ["result_two"]]
    except Exception as exc:
        assert False, f"Didn't find a call for nextset\n{exc}\n"
    else:
        assert True, "Found call for nextset"


@for_one_database
def test_nextset_with_two_select_statements_postgres(dsn, configuration):
    cursor = connect(dsn, **get_credentials(configuration)).cursor()
    cursor.execute("SELECT 4;SELECT 2;")
    try:
        assert cursor.fetchall() == [[4]]
        next_set_present = cursor.nextset()
        if next_set_present:
            assert cursor.fetchall() == [[2]]
    except Exception as exc:
        assert False, f"Didn't find a call for nextset\n{exc}\n"
    else:
        assert True, "Found call for nextset"


@for_each_database_except(["PostgreSQL", "MSSQL"])
def test_nextset_with_two_result_set_mysql(dsn, configuration):
    cursor = connect(dsn, **get_credentials(configuration)).cursor()
    multi_result_set_stored_proc = """
    CREATE PROCEDURE TEST_PROC_TWO_INTS()
    BEGIN

        SELECT 4;
        SELECT 2;

    END
    """
    cursor.execute(multi_result_set_stored_proc)
    cursor.execute("CALL TEST_PROC_TWO_INTS();")
    try:
        assert cursor.fetchall() == [[4]]
        next_set_present = cursor.nextset()
        if next_set_present:
            assert cursor.fetchall() == [[2]]
    except Exception as exc:
        assert False, f"Didn't find a call for nextset\n{exc}\n"
    else:
        assert True, "Found call for nextset"


@for_each_database_except(["PostgreSQL", "MSSQL"])
def test_nextset_with_two_result_set_with_alias_mysql(dsn, configuration):
    cursor = connect(dsn, **get_credentials(configuration)).cursor()
    multi_result_set_stored_proc = """
    CREATE PROCEDURE TEST_PROC_TWO_ALIAS_INTS()
    BEGIN

        SELECT 4 as Four;
        SELECT 2 as Two;

    END
    """
    cursor.execute(multi_result_set_stored_proc)
    cursor.execute("CALL TEST_PROC_TWO_ALIAS_INTS();")
    try:
        assert cursor.fetchall() == [[4]]
        next_set_present = cursor.nextset()
        if next_set_present:
            assert cursor.fetchall() == [[2]]
    except Exception as exc:
        assert False, f"Didn't find a call for nextset\n{exc}\n"
    else:
        assert True, "Found call for nextset"


@for_each_database_except(["PostgreSQL", "MSSQL"])
def test_nextset_with_three_result_set_mysql(dsn, configuration):
    cursor = connect(dsn, **get_credentials(configuration)).cursor()
    multi_result_set_stored_proc = """
    CREATE PROCEDURE TEST_PROC_THREE_INTS()
    BEGIN

        SELECT 4;
        SELECT 3;
        SELECT 2;

    END
    """
    cursor.execute(multi_result_set_stored_proc)
    cursor.execute("CALL TEST_PROC_THREE_INTS();")
    try:
        assert cursor.fetchall() == [[4]]
        next_set_present = cursor.nextset()
        if next_set_present:
            assert cursor.fetchall() == [[3]]
            next_set_present = cursor.nextset()
            if next_set_present:
                assert cursor.fetchall() == [[2]]
    except Exception as exc:
        assert False, f"Didn't find a call for nextset\n{exc}\n"
    else:
        assert True, "Found call for nextset"


@for_each_database_except(["PostgreSQL", "MySQL"])
def test_nextset_with_two_result_set_mssql(dsn, configuration):
    cursor = connect(dsn, **get_credentials(configuration)).cursor()
    multi_result_set_stored_proc = """CREATE PROCEDURE TEST_PROC
    AS
    SET NOCOUNT ON
    SELECT 4
    SELECT 2
    """
    cursor.execute(multi_result_set_stored_proc)
    cursor.execute("EXEC TEST_PROC;")
    try:
        assert cursor.fetchall() == [[4]]
        next_set_present = cursor.nextset()
        if next_set_present:
            assert cursor.fetchall() == [[2]]
    except Exception as exc:
        assert False, f"Didn't find a call for nextset\n{exc}\n"
    else:
        assert True, "Found call for nextset"


@for_each_database_except(["PostgreSQL", "MySQL"])
def test_nextset_with_two_result_set_with_alias_mssql(dsn, configuration):
    cursor = connect(dsn, **get_credentials(configuration)).cursor()
    multi_result_set_stored_proc = """CREATE PROCEDURE TEST_PROC
    AS
    SET NOCOUNT ON
    SELECT 4 as Four
    SELECT 2 as Two
    """
    cursor.execute(multi_result_set_stored_proc)
    cursor.execute("EXEC TEST_PROC;")
    try:
        assert cursor.fetchall() == [[4]]
        next_set_present = cursor.nextset()
        if next_set_present:
            assert cursor.fetchall() == [[2]]
    except Exception as exc:
        assert False, f"Didn't find a call for nextset\n{exc}\n"
    else:
        assert True, "Found call for nextset"


@for_each_database_except(["PostgreSQL", "MySQL"])
def test_nextset_with_three_result_set_mssql(dsn, configuration):
    cursor = connect(dsn, **get_credentials(configuration)).cursor()
    multi_result_set_stored_proc = """CREATE PROCEDURE TEST_PROC
    AS
    SET NOCOUNT ON
    SELECT 4
    SELECT 3
    SELECT 2
    """
    cursor.execute(multi_result_set_stored_proc)
    cursor.execute("EXEC TEST_PROC;")
    try:
        assert cursor.fetchall() == [[4]]
        next_set_present = cursor.nextset()
        if next_set_present:
            assert cursor.fetchall() == [[3]]
            next_set_present = cursor.nextset()
            if next_set_present:
                assert cursor.fetchall() == [[2]]
    except Exception as exc:
        assert False, f"Didn't find a call for nextset\n{exc}\n"
    else:
        assert True, "Found call for nextset"



import json
import os
from contextlib import contextmanager
from typing import Optional

import pytest

import turbodbc


def generate_microseconds_with_precision(digits):
    microseconds = 0
    for i in range(digits):
        microseconds = 10 * microseconds + i + 1
    for i in range(6 - digits):
        microseconds *= 10

    return microseconds


def _get_config_files():
    variable = "TURBODBC_TEST_CONFIGURATION_FILES"
    try:
        raw = os.environ[variable]
        file_names = raw.split(",")
        return [file_name.strip() for file_name in file_names]
    except KeyError:
        raise KeyError(
            "Please set the environment variable {} to specify the configuration files as a comma-separated list".format(
                variable
            )
        )


def get_credentials(configuration):
    if "user" in configuration:
        return {
            configuration["capabilities"]["connection_user_option"]: configuration[
                "user"
            ],
            configuration["capabilities"]["connection_password_option"]: configuration[
                "password"
            ],
        }
    else:
        return {}


def _load_configuration(file_name):
    with open(file_name) as f:
        return json.load(f)


def _get_configuration(file_name):
    conf = _load_configuration(file_name)
    return (conf["data_source_name"], conf)


def _get_configurations():
    return [_get_configuration(file_name) for file_name in _get_config_files()]


"""
Use this decorator to execute a test function once for each database configuration.

Please note the test function *must* take the parameters `dsn` and `configuration`,
and in that order.

Example:

@for_each_database
def test_important_stuff(dsn, configuration):
    assert 1 == 2
"""
for_each_database = pytest.mark.parametrize("dsn,configuration", _get_configurations())

"""
Use this decorator to execute a test function once for each database configuration
except for those databases configurations listed in the parameter.

Please note the test function *must* take the parameters `dsn` and `configuration`,
and in that order.

Example:

@for_each_database_except(["MySQL"])
def test_important_stuff(dsn, configuration):
    assert 1 == 2
"""


def for_each_database_except(exceptions):
    configurations = _get_configurations()
    return pytest.mark.parametrize(
        "dsn,configuration", [c for c in configurations if c[0] not in exceptions]
    )


"""
Use this decorator to execute a test function once for a single database configuration.

Please note the test function *must* take the parameters `dsn` and `configuration`,
and in that order.

Example:

@for_one_database
def test_important_stuff(dsn, configuration):
    assert 1 == 2
"""
for_one_database = pytest.mark.parametrize(
    "dsn,configuration", [_get_configuration(_get_config_files()[0])]
)


def for_specific_databases(db_filter: Optional[str] = None):
    """
    Use this decorator to execute a test function for a specific database configuration that contains the given filter.

    Please note the test function *must* take the parameters `dsn` and `configuration`,
    and in that order.

    Example:

    @for_specific_databases("mssql")
    def test_important_stuff_only_for_mssql(dsn, configuration):
        assert 1 == 2
    """
    filtered_config_files = [fn for fn in _get_config_files() if db_filter in fn]
    return pytest.mark.parametrize(
        "dsn,configuration", [_get_configuration(cf) for cf in filtered_config_files]
    )


@contextmanager
def open_connection(
    configuration, rows_to_buffer=None, parameter_sets_to_buffer=100, **turbodbc_options
):
    dsn = configuration["data_source_name"]
    prefer_unicode = configuration.get("prefer_unicode", False)
    read_buffer_size = (
        turbodbc.Rows(rows_to_buffer) if rows_to_buffer else turbodbc.Megabytes(1)
    )

    options = turbodbc.make_options(
        read_buffer_size=read_buffer_size,
        parameter_sets_to_buffer=parameter_sets_to_buffer,
        prefer_unicode=prefer_unicode,
        **turbodbc_options,
    )
    connection = turbodbc.connect(
        dsn, turbodbc_options=options, **get_credentials(configuration)
    )

    yield connection
    connection.close()


@contextmanager
def open_cursor(
    configuration, rows_to_buffer=None, parameter_sets_to_buffer=100, **turbodbc_options
):
    with open_connection(
        configuration, rows_to_buffer, parameter_sets_to_buffer, **turbodbc_options
    ) as connection:
        cursor = connection.cursor()
        yield cursor
        cursor.close()



import gc
import weakref

import pytest
from helpers import for_each_database, for_one_database, get_credentials, open_cursor
from query_fixture import query_fixture

from turbodbc import DatabaseError, Error, InterfaceError, connect


@for_one_database
def test_new_cursor_properties(dsn, configuration):
    connection = connect(dsn, **get_credentials(configuration))
    cursor = connection.cursor()

    # https://www.python.org/dev/peps/pep-0249/#rowcount
    assert cursor.rowcount == -1
    assert cursor.description is None
    assert cursor.arraysize == 1


@for_one_database
def test_closed_cursor_raises_when_used(dsn, configuration):
    connection = connect(dsn, **get_credentials(configuration))
    cursor = connection.cursor()

    cursor.close()

    with pytest.raises(InterfaceError):
        cursor.execute("SELECT 42")

    with pytest.raises(InterfaceError):
        cursor.executemany("SELECT 42")

    with pytest.raises(InterfaceError):
        cursor.executemanycolumns("SELECT 42", [])

    with pytest.raises(InterfaceError):
        cursor.fetchone()

    with pytest.raises(InterfaceError):
        cursor.fetchmany()

    with pytest.raises(InterfaceError):
        cursor.fetchall()

    with pytest.raises(InterfaceError):
        next(cursor)


@for_one_database
def test_closing_twice_does_not_raise(dsn, configuration):
    connection = connect(dsn, **get_credentials(configuration))
    cursor = connection.cursor()

    cursor.close()
    cursor.close()


@for_one_database
def test_open_cursor_without_result_set_raises(dsn, configuration):
    connection = connect(dsn, **get_credentials(configuration))
    cursor = connection.cursor()

    with pytest.raises(InterfaceError):
        cursor.fetchone()


@for_one_database
def test_setinputsizes_does_not_raise(dsn, configuration):
    """
    It is legal for setinputsizes() to do nothing, so anything except
    raising an exception is ok
    """
    cursor = connect(dsn, **get_credentials(configuration)).cursor()
    cursor.setinputsizes([10, 20])


@for_one_database
def test_setoutputsize_does_not_raise(dsn, configuration):
    """
    It is legal for setinputsizes() to do nothing, so anything except
    raising an exception is ok
    """
    cursor = connect(dsn, **get_credentials(configuration)).cursor()
    cursor.setoutputsize(1000, 42)  # with column
    cursor.setoutputsize(1000, column=42)  # with column
    cursor.setoutputsize(1000)  # without column


@for_one_database
def test_rowcount_is_reset_after_execute_raises(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "INSERT INTEGER") as table_name:
            cursor.execute(f"INSERT INTO {table_name} VALUES (?)", [42])
            assert cursor.rowcount == 1
            with pytest.raises(Error):
                cursor.execute("this is not even a valid SQL statement")
            assert cursor.rowcount == -1


@for_one_database
def test_rowcount_is_reset_after_executemany_raises(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "INSERT INTEGER") as table_name:
            cursor.execute(f"INSERT INTO {table_name} VALUES (?)", [42])
            assert cursor.rowcount == 1
            with pytest.raises(Error):
                cursor.executemany("this is not even a valid SQL statement")
            assert cursor.rowcount == -1


@for_one_database
def test_rowcount_is_reset_after_executemanycolumns_raises(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "INSERT INTEGER") as table_name:
            cursor.execute(f"INSERT INTO {table_name} VALUES (?)", [42])
            assert cursor.rowcount == 1
            with pytest.raises(Error):
                cursor.executemanycolumns("this is not even a valid SQL statement", [])
            assert cursor.rowcount == -1


@for_one_database
def test_connection_does_not_strongly_reference_cursors(dsn, configuration):
    connection = connect(dsn, **get_credentials(configuration))
    cursor = connection.cursor()

    cursor_ref = weakref.ref(cursor)
    del cursor
    gc.collect()
    assert cursor_ref() is None


@for_one_database
def test_pep343_with_statement(dsn, configuration):
    with connect(dsn, **get_credentials(configuration)) as connection:
        with connection.cursor() as cursor:
            cursor.execute("SELECT 42")

        # cursor should be closed
        with pytest.raises(InterfaceError):
            cursor.execute("SELECT 42")


@for_each_database
def test_insert_duplicate_uniquecol_raises(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(
            cursor, configuration, "INSERT DUPLICATE UNIQUECOL"
        ) as table_name:
            with pytest.raises(DatabaseError):
                cursor.execute(f"INSERT INTO {table_name} VALUES (1)")
                # some databases (e.g. exasol) report failure not in the execute statement above, but only
                # when closing the odbc handle, i.e. at cursor.close:
                cursor.close()



import turbodbc.data_types
from turbodbc import BINARY, DATETIME, NUMBER, ROWID, STRING

ALL_TYPE_CODES = [
    turbodbc.data_types._BOOLEAN_CODE,
    turbodbc.data_types._INTEGER_CODE,
    turbodbc.data_types._FLOATING_POINT_CODE,
    turbodbc.data_types._STRING_CODE,
    turbodbc.data_types._UNICODE_CODE,
    turbodbc.data_types._TIMESTAMP_CODE,
    turbodbc.data_types._DATE_CODE,
]

ALL_DATA_TYPES = [STRING, BINARY, NUMBER, DATETIME, ROWID]


def test_each_type_code_matches_one_data_type():
    for type_code in ALL_TYPE_CODES:
        matches = [type for type in ALL_DATA_TYPES if type_code == type]
        assert 1 == len(matches)


def test_each_type_code_mismatches_all_but_one_data_type():
    for type_code in ALL_TYPE_CODES:
        mismatches = [type for type in ALL_DATA_TYPES if type_code != type]
        expected = len(ALL_DATA_TYPES) - 1
        assert expected == len(mismatches)



import datetime
import gc
import weakref
from collections import OrderedDict

import pyarrow as pa
import pytest
from helpers import (
    for_each_database,
    for_one_database,
    generate_microseconds_with_precision,
    open_cursor,
)
from query_fixture import query_fixture

import turbodbc


def _fix_case(configuration, string):
    """
    some databases return column names in upper case
    """
    capabilities = configuration["capabilities"]
    if capabilities["reports_column_names_as_upper_case"]:
        return string.upper()
    else:
        return string


@for_one_database
def test_arrow_without_result_set_raises(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with pytest.raises(turbodbc.InterfaceError):
            cursor.fetchallarrow()


@for_each_database
def test_arrow_empty_column(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "INSERT INTEGER") as table_name:
            cursor.execute(f"SELECT a FROM {table_name}")
            result = cursor.fetchallarrow()
            assert isinstance(result, pa.Table)
            assert result.num_columns == 1
            assert result.num_rows == 0


@for_each_database
def test_arrow_reference_count(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "INSERT INTEGER") as table_name:
            cursor.execute(f"SELECT a FROM {table_name}")
            result = cursor.fetchallarrow()
            result_ref = weakref.ref(result)
            del result
            gc.collect()
            assert result_ref() is None


@for_each_database
def test_arrow_int_column(dsn, configuration):
    with open_cursor(configuration) as cursor:
        cursor.execute("SELECT 42 AS a")
        result = cursor.fetchallarrow()
        assert isinstance(result, pa.Table)
        assert result.num_columns == 1
        assert result.num_rows == 1
        assert result.schema[0].name == _fix_case(configuration, "a")
        assert str(result.column(0).type) == "int64"
        assert result.column(0).to_pylist() == [42]


@for_each_database
def test_arrow_int_column_adaptive(dsn, configuration):
    with open_cursor(configuration) as cursor:
        cursor.execute("SELECT 42 AS a")
        result = cursor.fetchallarrow(adaptive_integers=True)
        assert str(result.schema[0].type) == "int8"
        assert result.column(0).to_pylist() == [42]


@for_each_database
def test_arrow_double_column(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "SELECT DOUBLE") as query:
            cursor.execute(query)
            result = cursor.fetchallarrow()
            assert isinstance(result, pa.Table)
            assert result.num_columns == 1
            assert result.num_rows == 1
            assert result.schema[0].name == _fix_case(configuration, "a")
            assert str(result.schema[0].type) == "double"
            assert result.column(0).to_pylist() == [3.14]


@for_each_database
def test_arrow_boolean_column(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "INSERT INDEXED BOOL") as table_name:
            cursor.executemany(
                f"INSERT INTO {table_name} VALUES (?, ?)",
                [[True, 1], [False, 2], [True, 3]],
            )
            cursor.execute(f"SELECT a FROM {table_name} ORDER BY b")
            result = cursor.fetchallarrow()
            assert isinstance(result, pa.Table)
            assert result.num_columns == 1
            assert result.num_rows == 3
            assert result.schema[0].name == _fix_case(configuration, "a")
            assert str(result.schema[0].type) == "bool"
            assert result.column(0).to_pylist() == [True, False, True]


@for_each_database
def test_arrow_binary_column_with_null(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(
            cursor, configuration, "INSERT TWO INTEGER COLUMNS"
        ) as table_name:
            cursor.executemany(
                f"INSERT INTO {table_name} VALUES (?, ?)", [[42, 1], [None, 2]]
            )  # second column to enforce ordering
            cursor.execute(f"SELECT a FROM {table_name} ORDER BY b")
            result = cursor.fetchallarrow()
            assert isinstance(result, pa.Table)
            assert result.num_columns == 1
            assert result.num_rows == 2
            assert result.schema[0].name == _fix_case(configuration, "a")
            assert str(result.schema[0].type) == "int64"
            assert result.column(0).to_pylist() == [42, None]
            assert result.column(0).null_count == 1


@for_each_database
def test_arrow_binary_column_larger_than_batch_size(dsn, configuration):
    with open_cursor(configuration, rows_to_buffer=2) as cursor:
        with query_fixture(cursor, configuration, "INSERT INTEGER") as table_name:
            cursor.executemany(
                f"INSERT INTO {table_name} VALUES (?)",
                [[1], [2], [3], [4], [5]],
            )
            cursor.execute(f"SELECT a FROM {table_name} ORDER BY a")
            result = cursor.fetchallarrow()
            assert isinstance(result, pa.Table)
            assert result.column(0).to_pylist() == [1, 2, 3, 4, 5]


@for_each_database
def test_arrow_timestamp_column(dsn, configuration):
    supported_digits = configuration["capabilities"]["fractional_second_digits"]
    fractional = generate_microseconds_with_precision(supported_digits)
    timestamp = datetime.datetime(2015, 12, 31, 1, 2, 3, fractional)

    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "INSERT TIMESTAMP") as table_name:
            cursor.execute(f"INSERT INTO {table_name} VALUES (?)", [timestamp])
            cursor.execute(f"SELECT a FROM {table_name}")
            result = cursor.fetchallarrow()
            assert result.column(0).to_pylist() == [timestamp]


@for_each_database
def test_arrow_date_column(dsn, configuration):
    date = datetime.date(2015, 12, 31)

    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "INSERT DATE") as table_name:
            cursor.execute(f"INSERT INTO {table_name} VALUES (?)", [date])
            cursor.execute(f"SELECT a FROM {table_name}")
            result = cursor.fetchallarrow()
            result.column(0).to_pylist() == [datetime.date(2015, 12, 31)]


@for_each_database
def test_arrow_timelike_column_with_null(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "INSERT TIMESTAMP") as table_name:
            cursor.execute(f"INSERT INTO {table_name} VALUES (?)", [None])
            cursor.execute(f"SELECT a FROM {table_name}")
            result = cursor.fetchallarrow()
            assert result.column(0).to_pylist() == [None]


@for_each_database
def test_arrow_timelike_column_larger_than_batch_size(dsn, configuration):
    timestamps = [
        datetime.datetime(2015, 12, 31, 1, 2, 3),
        datetime.datetime(2016, 1, 5, 4, 5, 6),
        datetime.datetime(2017, 2, 6, 7, 8, 9),
        datetime.datetime(2018, 3, 7, 10, 11, 12),
        datetime.datetime(2019, 4, 8, 13, 14, 15),
    ]

    with open_cursor(configuration, rows_to_buffer=2) as cursor:
        with query_fixture(cursor, configuration, "INSERT TIMESTAMP") as table_name:
            cursor.executemany(
                f"INSERT INTO {table_name} VALUES (?)",
                [[timestamp] for timestamp in timestamps],
            )
            cursor.execute(f"SELECT a FROM {table_name} ORDER BY a")
            result = cursor.fetchallarrow()
            assert result.column(0).to_pylist() == timestamps


@for_each_database
@pytest.mark.parametrize("strings_as_dictionary", [True, False])
def test_arrow_string_column(dsn, configuration, strings_as_dictionary):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "INSERT UNICODE") as table_name:
            cursor.execute(f"INSERT INTO {table_name} VALUES (?)", ["unicode \u2665"])
            cursor.execute(f"SELECT a FROM {table_name}")
            result = cursor.fetchallarrow(strings_as_dictionary=strings_as_dictionary)
            assert result.column(0).to_pylist() == ["unicode \u2665"]


@for_each_database
@pytest.mark.parametrize("strings_as_dictionary", [True, False])
def test_arrow_string_column_with_null(dsn, configuration, strings_as_dictionary):
    with open_cursor(configuration) as cursor:
        with query_fixture(cursor, configuration, "INSERT STRING") as table_name:
            cursor.execute(f"INSERT INTO {table_name} VALUES (?)", [None])
            cursor.execute(f"SELECT a FROM {table_name}")
            result = cursor.fetchallarrow(strings_as_dictionary=strings_as_dictionary)
            result.column(0).null_count == 1
            result.column(0).to_pylist() == [None]


@for_each_database
@pytest.mark.parametrize("strings_as_dictionary", [True, False])
def test_arrow_string_column_larger_than_batch_size(
    dsn, configuration, strings_as_dictionary
):
    strings = ["abc", "def", "ghi", "jkl", "mno"]
    with open_cursor(configuration, rows_to_buffer=2) as cursor:
        with query_fixture(cursor, configuration, "INSERT STRING") as table_name:
            cursor.executemany(
                f"INSERT INTO {table_name} VALUES (?)",
                [[string] for string in strings],
            )
            cursor.execute(f"SELECT a FROM {table_name} ORDER BY a")
            result = cursor.fetchallarrow(strings_as_dictionary=strings_as_dictionary)
            result.column(0).to_pylist() == strings


@for_each_database
def test_arrow_two_columns(dsn, configuration):
    with open_cursor(configuration) as cursor:
        with query_fixture(
            cursor, configuration, "INSERT TWO INTEGER COLUMNS"
        ) as table_name:
            cursor.executemany(
                f"INSERT INTO {table_name} VALUES (?, ?)", [[1, 42], [2, 41]]
            )
            cursor.execute(f"SELECT a, b FROM {table_name} ORDER BY a")
            result = cursor.fetchallarrow()
            assert result.to_pydict() == OrderedDict(
                [
                    (_fix_case(configuration, "a"), [1, 2]),
                    (_fix_case(configuration, "b"), [42, 41]),
                ]
            )


@for_each_database
def test_arrow_two_columns_single_row_buffer(dsn, configuration):
    with open_cursor(configuration, rows_to_buffer=1) as cursor:
        with query_fixture(
            cursor, configuration, "INSERT TWO INTEGER COLUMNS"
        ) as table_name:
            cursor.executemany(
                f"INSERT INTO {table_name} VALUES (?, ?)", [[1, 42], [2, 41]]
            )
            cursor.execute(f"SELECT a, b FROM {table_name} ORDER BY a")
            result = list(cursor.fetcharrowbatches())
            assert len(result) == 2



from collections import OrderedDict
from itertools import islice
from typing import Any, cast

import pyarrow as pa
from turbodbc_intern import (
    make_arrow_result_set,
    make_numpy_result_set,
    make_parameter_set,
    make_row_based_result_set,
    set_arrow_parameters,
    set_numpy_parameters,
)

from .exceptions import InterfaceError, translate_exceptions


def _make_masked_arrays(result_batch):
    from numpy import object_
    from numpy.ma import MaskedArray

    masked_arrays = []
    for data, mask in result_batch:
        if isinstance(data, list):
            masked_arrays.append(MaskedArray(data=data, mask=mask, dtype=object_))
        else:
            masked_arrays.append(MaskedArray(data=data, mask=mask))
    return masked_arrays


def _assert_numpy_column_preconditions(columns):
    from numpy import ndarray
    from numpy.ma import MaskedArray

    n_columns = len(columns)
    for index, column in enumerate(columns, start=1):
        if type(column) not in [MaskedArray, ndarray]:
            raise InterfaceError(
                "Bad type for column {} of {}. Only numpy.ndarray and numpy.ma.MaskedArrays are supported".format(
                    index, n_columns
                )
            )
        if column.ndim != 1:
            raise InterfaceError(
                f"Column {index} of {n_columns} is not one-dimensional"
            )
        if not column.flags.c_contiguous:
            raise InterfaceError(f"Column {index} of {n_columns} is not contiguous")

    lengths = [len(column) for column in columns]
    all_same_length = all(length == lengths[0] for length in lengths)
    if not all_same_length:
        raise InterfaceError(
            f"All columns must have the same length, got lengths {lengths}"
        )


class Cursor:
    """
    This class allows you to send SQL commands and queries to a database and retrieve
    associated result sets.
    """

    def __init__(self, impl):
        self.impl = impl
        self.result_set = None
        self.rowcount = -1
        self.arraysize = 1

    def __iter__(self):
        return self

    def __next__(self):
        element = self.fetchone()
        if element is None:
            raise StopIteration
        else:
            return element

    def _assert_valid(self):
        if self.impl is None:
            raise InterfaceError("Cursor already closed")

    def _assert_valid_result_set(self):
        if self.result_set is None:
            raise InterfaceError("No active result set")

    @property
    def description(self):
        """
        Retrieve a description of the columns in the current result set

        :return: A tuple of seven elements. Only some elements are meaningful:\n
                 *   Element #0 is the name of the column
                 *   Element #1 is the type code of the column
                 *   Element #6 is true if the column may contain ``NULL`` values
        """
        if self.result_set:
            info = self.result_set.get_column_info()
            return [
                (c.name, c.type_code(), None, None, None, None, c.supports_null_values)
                for c in info
            ]
        else:
            return None

    def _execute(self):
        self.impl.execute()
        self.rowcount = self.impl.get_row_count()
        cpp_result_set = self.impl.get_result_set()
        if cpp_result_set:
            self.result_set = make_row_based_result_set(cpp_result_set)
        else:
            self.result_set = None
        return self

    @translate_exceptions
    def execute(self, sql, parameters=None):
        """
        Execute an SQL command or query

        :param sql: A (unicode) string that contains the SQL command or query. If you would like to
               use parameters, please use a question mark ``?`` at the location where the
               parameter shall be inserted.
        :param parameters: An iterable of parameter values. The number of values must match
               the number of parameters in the SQL string.
        :return: The ``Cursor`` object to allow chaining of operations.
        """
        self.rowcount = -1
        self._assert_valid()
        self.impl.prepare(sql)
        if parameters:
            buffer = make_parameter_set(self.impl)
            buffer.add_set(parameters)
            buffer.flush()
        return self._execute()

    @translate_exceptions
    def executemany(self, sql, parameters=None):
        """
        Execute an SQL command or query with multiple parameter sets passed in a row-wise fashion.
        This function is part of PEP-249.

        :param sql: A (unicode) string that contains the SQL command or query. If you would like to
               use parameters, please use a question mark ``?`` at the location where the
               parameter shall be inserted.
        :param parameters: An iterable of iterable of parameter values. The outer iterable represents
               separate parameter sets. The inner iterable contains parameter values for a given
               parameter set. The number of values of each set must match the number of parameters
               in the SQL string.
        :return: The ``Cursor`` object to allow chaining of operations.
        """
        self.rowcount = -1
        self._assert_valid()
        self.impl.prepare(sql)

        if parameters:
            buffer = make_parameter_set(self.impl)
            for parameter_set in parameters:
                buffer.add_set(parameter_set)
            buffer.flush()

        return self._execute()

    @translate_exceptions
    def executemanycolumns(self, sql: str, columns):
        """
        Execute an SQL command or query with multiple parameter sets that are passed in
        a column-wise fashion as opposed to the row-wise parameters in ``executemany()``.
        This function is a turbodbc-specific extension to PEP-249.

        :param sql: A (unicode) string that contains the SQL command or query. If you would like to
               use parameters, please use a question mark ``?`` at the location where the
               parameter shall be inserted.
        :param columns: An iterable of NumPy MaskedArrays. The Arrays represent the columnar
               parameter data,
        :return: The ``Cursor`` object to allow chaining of operations.
        """
        self.rowcount = -1
        self._assert_valid()

        self.impl.prepare(sql)

        def _num_chunks(c):
            return c.num_chunks

        if isinstance(columns, pa.Table):
            for column in columns.itercolumns():
                if _num_chunks(column) != 1:
                    raise NotImplementedError("Chunked Arrays are " "not yet supported")

            set_arrow_parameters(self.impl, columns)
            return self._execute()

        _assert_numpy_column_preconditions(columns)

        from numpy.ma import MaskedArray

        split_arrays = []
        for column in columns:
            if isinstance(column, MaskedArray):
                split_arrays.append((column.data, column.mask, str(column.dtype)))
            else:
                split_arrays.append((column, False, str(column.dtype)))
        set_numpy_parameters(self.impl, split_arrays)

        return self._execute()

    @translate_exceptions
    def fetchone(self):
        """
        Returns a single row of a result set. Requires an active result set on the database
        generated with ``execute()`` or ``executemany()``.

        :return: Returns ``None`` when no more rows are available in the result set
        """
        self._assert_valid_result_set()
        result = cast(Any, self.result_set).fetch_row()
        if len(result) == 0:
            return None
        else:
            return result

    @translate_exceptions
    def fetchall(self):
        """
        Fetches a list of all rows in the active result set generated with ``execute()`` or
        ``executemany()``.

        :return: A list of rows
        """
        return [row for row in self]

    @translate_exceptions
    def fetchmany(self, size=None):
        """
        Fetches a batch of rows in the active result set generated with ``execute()`` or
        ``executemany()``.

        :param size: Controls how many rows are returned. The default ``None`` means that
               the value of Cursor.arraysize is used.
        :return: A list of rows
        """
        if size is None:
            size = self.arraysize
        if size <= 0:
            raise InterfaceError(f"Invalid arraysize {size} for fetchmany()")

        return [row for row in islice(self, size)]

    def fetchallnumpy(self):
        """
        Fetches all rows in the active result set generated with ``execute()`` or
        ``executemany()``.

        :return: An ``OrderedDict`` of *columns*, where the keys of the dictionary
                 are the column names. The columns are of NumPy's ``MaskedArray``
                 type, where the optimal data type for each result set column is
                 chosen automatically.
        """
        from numpy.ma import concatenate

        batches = list(self._numpy_batch_generator())
        column_names = [description[0] for description in self.description]
        return OrderedDict(
            zip(column_names, [concatenate(column) for column in zip(*batches)])
        )

    def fetchnumpybatches(self):
        """
        Returns an iterator over all rows in the active result set generated with ``execute()`` or
        ``executemany()``.

        :return: An iterator you can use to iterate over batches of rows of the result set. Each
                 batch consists of an ``OrderedDict`` of NumPy ``MaskedArray`` instances. See
                 ``fetchallnumpy()`` for details.
        """
        batchgen = self._numpy_batch_generator()
        column_names = [description[0] for description in self.description]
        for next_batch in batchgen:
            yield OrderedDict(zip(column_names, next_batch))

    def _numpy_batch_generator(self):
        self._assert_valid_result_set()

        numpy_result_set = make_numpy_result_set(self.impl.get_result_set())
        first_run = True
        while True:
            result_batch = _make_masked_arrays(numpy_result_set.fetch_next_batch())
            is_empty_batch = len(result_batch[0]) == 0
            if is_empty_batch and not first_run:
                return  # Let us return a typed result set at least once
            first_run = False
            yield result_batch

    def fetcharrowbatches(self, strings_as_dictionary=False, adaptive_integers=False):
        """
        Fetches rows in the active result set generated with ``execute()`` or
        ``executemany()`` as an iterable of arrow tables.

        :param strings_as_dictionary: If true, fetch string columns as
                 dictionary[string] instead of a plain string column.

        :param adaptive_integers: If true, instead of the integer type returned
                by the database (driver), this produce integer columns with the
                smallest possible integer type in which all values can be
                stored. Be aware that here the type depends on the resulting
                data.

        :return: generator of ``pyarrow.Table``
        """
        self._assert_valid_result_set()

        rs = make_arrow_result_set(
            self.impl.get_result_set(), strings_as_dictionary, adaptive_integers
        )
        first_run = True
        while True:
            table = rs.fetch_next_batch()
            is_empty_batch = len(table) == 0
            if is_empty_batch and not first_run:
                return  # Let us return a typed result set at least once
            first_run = False
            yield table

    def fetchallarrow(self, strings_as_dictionary=False, adaptive_integers=False):
        """
        Fetches all rows in the active result set generated with ``execute()`` or
        ``executemany()``.

        :param strings_as_dictionary: If true, fetch string columns as
                 dictionary[string] instead of a plain string column.

        :param adaptive_integers: If true, instead of the integer type returned
                by the database (driver), this produce integer columns with the
                smallest possible integer type in which all values can be
                stored. Be aware that here the type depends on the resulting
                data.

        :return: ``pyarrow.Table``
        """
        self._assert_valid_result_set()

        return make_arrow_result_set(
            self.impl.get_result_set(), strings_as_dictionary, adaptive_integers
        ).fetch_all()

    def nextset(self) -> bool:
        """
        makes cursor skip to the next available result set, discarding any remaining rows from the
        current result set. If there are no more result sets, the method returns False.
        Otherwise, it returns a True and subsequent calls to the fetch methods will
        return rows from the next result set.
        """
        self._assert_valid_result_set()
        return self.impl.more_results()

    def close(self):
        """
        Close the cursor.
        """
        self.result_set = None
        if self.impl is not None:
            self.impl._reset()
        self.impl = None

    def setinputsizes(self, sizes):
        """
        Has no effect since turbodbc automatically picks appropriate
        return types and sizes. Method exists since PEP-249 requires it.
        """
        pass

    def setoutputsize(self, size, column=None):
        """
        Has no effect since turbodbc automatically picks appropriate
        input types and sizes. Method exists since PEP-249 requires it.
        """
        pass

    def __enter__(self):
        """
        Conformance to PEP-343
        """
        return self

    def __exit__(self, type, value, traceback):
        """
        Conformance to PEP-343
        """
        return self.close()
